
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_instance_metadata.c                                    */
/*  DESCRIPTION:  Instance metadata                                          */
/*                                                                           */
/*****************************************************************************/
#include <stdarg.h>
#include "khe_interns.h"


/*****************************************************************************/
/*                                                                           */
/*  KHE_INSTANCE_METADATA - instance metadata                                */
/*                                                                           */
/*****************************************************************************/

struct khe_instance_metadata_rec {
  char			*name;			/* instance name             */
  char			*contributor;		/* contributor               */
  char			*date;			/* date                      */
  char			*country;		/* country                   */
  char			*description;		/* description               */
  char			*remarks;		/* remarks (optional)        */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "construction and query"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_INSTANCE_METADATA KheInstanceMetaDataMake(char *name, char *contr,   */
/*    char *date, char *country, char *description, char *remarks)           */
/*                                                                           */
/*  Make instance metadata with these attributes.                            */
/*                                                                           */
/*  Implementation note.  MStringCopy accepts value NULL and returns NULL.   */
/*                                                                           */
/*****************************************************************************/

KHE_INSTANCE_METADATA KheInstanceMetaDataMake(char *name, char *contributor,
  char *date, char *country, char *description, char *remarks)
{
  KHE_INSTANCE_METADATA res;
  MMake(res);
  res->name = MStringCopy(name);
  res->contributor = MStringCopy(contributor);
  res->date = MStringCopy(date);
  res->country = MStringCopy(country);
  res->description = MStringCopy(description);
  res->remarks = MStringCopy(remarks);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheInstanceMetaDataName(KHE_INSTANCE_METADATA md)                  */
/*                                                                           */
/*  Return the Name attribute of md.                                         */
/*                                                                           */
/*****************************************************************************/

char *KheInstanceMetaDataName(KHE_INSTANCE_METADATA md)
{
  return md->name;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheInstanceMetaDataContributor(KHE_INSTANCE_METADATA md)           */
/*                                                                           */
/*  Return the Contributor attribute of md.                                  */
/*                                                                           */
/*****************************************************************************/

char *KheInstanceMetaDataContributor(KHE_INSTANCE_METADATA md)
{
  return md->contributor;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheInstanceMetaDataDate(KHE_INSTANCE_METADATA md)                  */
/*                                                                           */
/*  Return the Date attribute of md.                                         */
/*                                                                           */
/*****************************************************************************/

char *KheInstanceMetaDataDate(KHE_INSTANCE_METADATA md)
{
  return md->date;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheInstanceMetaDataCountry(KHE_INSTANCE_METADATA md)               */
/*                                                                           */
/*  Return the Country attribute of md.                                      */
/*                                                                           */
/*****************************************************************************/

char *KheInstanceMetaDataCountry(KHE_INSTANCE_METADATA md)
{
  return md->country;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheInstanceMetaDataDescription(KHE_INSTANCE_METADATA md)           */
/*                                                                           */
/*  Return the Description attribute of md.                                  */
/*                                                                           */
/*****************************************************************************/

char *KheInstanceMetaDataDescription(KHE_INSTANCE_METADATA md)
{
  return md->description;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheInstanceMetaDataRemarks(KHE_INSTANCE_METADATA md)               */
/*                                                                           */
/*  Return the optional Remarks attribute of md.                             */
/*                                                                           */
/*****************************************************************************/

char *KheInstanceMetaDataRemarks(KHE_INSTANCE_METADATA md)
{
  return md->remarks;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "reading and writing"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheInstanceMetaDataMakeFromKml(KML_ELT md_elt, KHE_INSTANCE ins,    */
/*    KML_ERROR *ke)                                                         */
/*                                                                           */
/*  Make metadata from md_elt and add it to ins.                             */
/*                                                                           */
/*****************************************************************************/

bool KheInstanceMetaDataMakeFromKml(KML_ELT md_elt, KHE_INSTANCE ins,
  KML_ERROR *ke)
{
  KHE_INSTANCE_METADATA res;
  if( !KmlCheck(md_elt,
      ": $Name $Contributor $Date $Country $Description +$Remarks", ke) )
    return false;
  res = KheInstanceMetaDataMake(KmlText(KmlChild(md_elt, 0)),
    KmlText(KmlChild(md_elt, 1)), KmlText(KmlChild(md_elt, 2)),
    KmlText(KmlChild(md_elt, 3)), KmlText(KmlChild(md_elt, 4)),
    KmlChildCount(md_elt) == 6 ? KmlText(KmlChild(md_elt, 5)) : NULL);
  KheInstanceSetMetaData(ins, res);
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheInstanceMetaDataWrite(KHE_INSTANCE_METADATA md, KML_FILE kf)     */
/*                                                                           */
/*  Write instance metadata to xml.  Here md may be NULL, in which case      */
/*  a placeholder value will be written.                                     */
/*                                                                           */
/*****************************************************************************/

static bool Present(char *val)
{
  return val != NULL && *val != '\0';
}

static char *Val(char *val, char *alt)
{
  return Present(val) ? val : alt;
}

void KheInstanceMetaDataWrite(KHE_INSTANCE_METADATA md, KML_FILE kf)
{
  KmlBegin(kf, "MetaData");
  if( md != NULL )
  {
    KmlEltPlainText(kf, "Name", Val(md->name, "No name"));
    KmlEltPlainText(kf, "Contributor", Val(md->contributor, "No contributor"));
    KmlEltPlainText(kf, "Date", Val(md->date, "No date"));
    KmlEltPlainText(kf, "Country", Val(md->country, "No country"));
    KmlEltPlainText(kf, "Description", Val(md->description, "No description"));
    if( Present(md->remarks) )
      KmlEltPlainText(kf, "Remarks", md->remarks);
  }
  else
  {
    KmlEltPlainText(kf, "Name", "No name");
    KmlEltPlainText(kf, "Contributor", "No contributor");
    KmlEltPlainText(kf, "Date", "No date");
    KmlEltPlainText(kf, "Country", "No country");
    KmlEltPlainText(kf, "Description", "No description");
  }
  KmlEnd(kf, "MetaData");
}
