
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_archive.c                                              */
/*  DESCRIPTION:  An archive holding instances and solution groups           */
/*                                                                           */
/*****************************************************************************/
#include "khe_interns.h"
#include <stdarg.h>
#define MAX_ERROR_STRING 200
#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0


/*****************************************************************************/
/*                                                                           */
/*  KHE_INDEXED_INSTANCE - an instance with its index in archive (private)   */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_indexed_instance_rec {
  KHE_INSTANCE			instance;
  int				index;
} *KHE_INDEXED_INSTANCE;

typedef HA_ARRAY(KHE_INDEXED_INSTANCE) ARRAY_KHE_INDEXED_INSTANCE;
typedef HN_TABLE(KHE_INDEXED_INSTANCE) TABLE_KHE_INDEXED_INSTANCE;


/*****************************************************************************/
/*                                                                           */
/*  KHE_ARCHIVE - one archive of problems and solution groups                */
/*                                                                           */
/*****************************************************************************/

struct khe_archive_rec {
  HA_ARENA			arena;			/* arena for archive */
  void				*back;			/* back pointer      */
  char				*id;			/* optional Id       */
  KHE_ARCHIVE_METADATA		meta_data;		/* optional MetaData */
  KHE_MODEL			model;			/* high school etc.  */
  ARRAY_KHE_INDEXED_INSTANCE	instance_array;		/* instance array    */
  TABLE_KHE_INDEXED_INSTANCE	instance_table;		/* instance table    */
  ARRAY_KHE_SOLN_GROUP		soln_group_array;	/* solution groups   */
  TABLE_KHE_SOLN_GROUP		soln_group_table;	/* solution groups   */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "indexed instances"                                            */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_INDEXED_INSTANCE KheIndexedInstanceMake(KHE_INSTANCE ins, int index) */
/*                                                                           */
/*  Make a new indexed instance object with these attributes.                */
/*                                                                           */
/*****************************************************************************/

static KHE_INDEXED_INSTANCE KheIndexedInstanceMake(KHE_INSTANCE ins, int index,
  HA_ARENA a)
{
  KHE_INDEXED_INSTANCE res;
  HaMake(res, a);
  res->instance = ins;
  res->index = index;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheIndexedInstanceDelete(KHE_INDEXED_INSTANCE ii)                   */
/*                                                                           */
/*  Delete ii.                                                               */
/*                                                                           */
/*****************************************************************************/

/* *** don't delete now that there are arenas
static void KheIndexedInstanceDelete(KHE_INDEXED_INSTANCE ii)
{
  MFree(ii);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "construction and query"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_ARCHIVE KheArchiveMake(char *id, KHE_ARCHIVE_METADATA md,            */
/*    KHE_MODEL model)                                                       */
/*                                                                           */
/*  Make an initially empty archive, in its own arena.                       */
/*                                                                           */
/*****************************************************************************/

KHE_ARCHIVE KheArchiveMake(char *id, KHE_ARCHIVE_METADATA md, KHE_MODEL model)
{
  KHE_ARCHIVE res;  HA_ARENA a;
  a = HaArenaMake();
  HaMake(res, a);
  res->arena = a;
  res->back = NULL;
  res->id = HnStringCopy(id, a);
  res->meta_data = md;
  res->model = model;
  HaArrayInit(res->instance_array, a);
  HnTableInit(res->instance_table, a);
  HaArrayInit(res->soln_group_array, a);
  HnTableInit(res->soln_group_table, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheArchiveId(KHE_ARCHIVE archive)                                  */
/*                                                                           */
/*  Return the Id of archive, possibly NULL.                                 */
/*                                                                           */
/*****************************************************************************/

char *KheArchiveId(KHE_ARCHIVE archive)
{
  return archive->id;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveSetBack(KHE_ARCHIVE archive, void *back)                  */
/*                                                                           */
/*  Set the back pointer of archive.                                         */
/*                                                                           */
/*****************************************************************************/

void KheArchiveSetBack(KHE_ARCHIVE archive, void *back)
{
  archive->back = back;
}


/*****************************************************************************/
/*                                                                           */
/*  void *KheArchiveBack(KHE_ARCHIVE archive)                                */
/*                                                                           */
/*  Retrieve the back pointer of archive.                                    */
/*                                                                           */
/*****************************************************************************/

void *KheArchiveBack(KHE_ARCHIVE archive)
{
  return archive->back;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_ARCHIVE_METADATA KheArchiveMetaData(KHE_ARCHIVE archive)             */
/*                                                                           */
/*  Return the MetaData attribute of archive.                                */
/*                                                                           */
/*****************************************************************************/

KHE_ARCHIVE_METADATA KheArchiveMetaData(KHE_ARCHIVE archive)
{
  return archive->meta_data;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveSetMetaData(KHE_ARCHIVE archive, KHE_ARCHIVE_METADATA md) */
/*                                                                           */
/*  Set the metadata attribute of archive to md.                             */
/*                                                                           */
/*****************************************************************************/

void KheArchiveSetMetaData(KHE_ARCHIVE archive, KHE_ARCHIVE_METADATA md)
{
  archive->meta_data = md;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_MODEL KheArchiveModel(KHE_ARCHIVE archive)                           */
/*                                                                           */
/*  Return the model of archive.                                             */
/*                                                                           */
/*****************************************************************************/

KHE_MODEL KheArchiveModel(KHE_ARCHIVE archive)
{
  return archive->model;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "instances"                                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveAddInstance(KHE_ARCHIVE archive, KHE_INSTANCE ins)        */
/*                                                                           */
/*  Add ins to archive, assuming it is safe to do so (not duplicate Id).     */
/*                                                                           */
/*****************************************************************************/

bool KheArchiveAddInstance(KHE_ARCHIVE archive, KHE_INSTANCE ins)
{
  KHE_INDEXED_INSTANCE ii;  char *id;  int pos;

  /* make sure models agree */
  if( KheInstanceModel(ins) != KheArchiveModel(archive) )
    return false;

  /* if the instance has an id, make sure it is not already in use */
  id = KheInstanceId(ins);
  if( id != NULL && HnTableContains(archive->instance_table, id, pos) )
    return false;

  /* all good, so go ahead and add the instance */
  ii = KheIndexedInstanceMake(ins, HaArrayCount(archive->instance_array),
    archive->arena);
  HaArrayAddLast(archive->instance_array, ii);
  if( id != NULL )
    HnTableAdd(archive->instance_table, id, ii);
  KheInstanceAddArchive(ins, archive);
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveDeleteInstance(KHE_ARCHIVE archive, KHE_INSTANCE ins)     */
/*                                                                           */
/*  Delete ins from archive.  Also delete any solutions for ins.             */
/*                                                                           */
/*****************************************************************************/

void KheArchiveDeleteInstance(KHE_ARCHIVE archive, KHE_INSTANCE ins)
{
  int pos, i;  char *id;  KHE_SOLN_GROUP sg;
  KHE_INDEXED_INSTANCE ii, ii2;

  /* find ins's indexed instance, either by retrieving by id or searching; */
  /* also delete the indexed instance from archive's table */
  ii = NULL;
  id = KheInstanceId(ins);
  if( id != NULL )
  {
    if( !HnTableRetrieve(archive->instance_table, id, ii, pos) )
      HnAbort("KheArchiveDeleteInstance: instance %s not in archive", id);
    HnTableDelete(archive->instance_table, pos);
  }
  else
  {
    HaArrayForEach(archive->instance_array, ii, i)
      if( ii->instance == ins )
	break;
    if( i >= HaArrayCount(archive->instance_array) )
      HnAbort("KheArchiveDeleteInstance: instance not in archive");
  }

  /* remove ii from archive's array and change indexes of following entries */
  ii2 = HaArray(archive->instance_array, ii->index);
  HnAssert(ii2 == ii && ii->instance == ins,
    "KheArchiveDeleteInstance internal error");
  for( i = ii->index + 1;  i < HaArrayCount(archive->instance_array);  i++ )
  {
    ii2 = HaArray(archive->instance_array, i);
    HaArrayPut(archive->instance_array, i - 1, ii2);
    ii2->index = i - 1;
  }
  HaArrayDeleteLast(archive->instance_array);

  /* delete solutions for ins from archive's solution groups */
  HaArrayForEach(archive->soln_group_array, sg, i)
    KheSolnGroupDeleteInstance(sg, ii->instance, ii->index);

  /* delete ii */
  /* KheIndexedInstanceDelete(ii); */

  /* delete the link between archive and ins */
  /* ***
  HnAssert(MArrayContains(archive->instance_array, ins, &pos),
    "KheArchiveDeleteInstance: ins not in archive");
  MArrayRemove(archive->instance_array, pos);
  KheInstanceDeleteArchive(ins, archive);
  if( KheInstanceId(ins) != NULL )
  {
    if( !MTableContains(archive->instance_table, KheInstanceId(ins), &pos) )
      HnAbort("KheArchiveDeleteInstance: internal error");
    MTableDelete(archive->instance_table, pos);
  }
  *** */

  /* delete solutions for ins in archive */
  /* ***
  HaArrayForEach(archive->soln_group_array, sg, i)
    KheSolnGroupDeleteSolnsForInstance(sg, ins);
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  int KheArchiveInstanceCount(KHE_ARCHIVE archive)                         */
/*                                                                           */
/*  Return the number of instances in archive.                               */
/*                                                                           */
/*****************************************************************************/

int KheArchiveInstanceCount(KHE_ARCHIVE archive)
{
  return HaArrayCount(archive->instance_array);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_INSTANCE KheArchiveInstance(KHE_ARCHIVE archive, int i)              */
/*                                                                           */
/*  Return the i'th instance of archive.                                     */
/*                                                                           */
/*****************************************************************************/

KHE_INSTANCE KheArchiveInstance(KHE_ARCHIVE archive, int i)
{
  return HaArray(archive->instance_array, i)->instance;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveRetrieveInstance(KHE_ARCHIVE archive, char *id,           */
/*    KHE_INSTANCE *ins)                                                     */
/*                                                                           */
/*  Retrieve an instance with the given id from archive.                     */
/*                                                                           */
/*****************************************************************************/

bool KheArchiveRetrieveInstance(KHE_ARCHIVE archive, char *id,
  KHE_INSTANCE *ins, int *index)
{
  int pos;  KHE_INDEXED_INSTANCE ii;
  if( HnTableRetrieve(archive->instance_table, id, ii, pos) )
    return *ins = ii->instance, *index = ii->index, true;
  else
    return *ins = NULL, *index = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveFindInstance(KHE_ARCHIVE archive, KHE_INSTANCE ins,       */
/*    KHE_INDEXED_INSTANCE *ii)                                              */
/*                                                                           */
/*  If archive contains ins, return true with *ii set to its indexed         */
/*  instance object.  Otherwise return false, with *ii set to NULL.          */
/*                                                                           */
/*****************************************************************************/

static bool KheArchiveFindInstance(KHE_ARCHIVE archive, KHE_INSTANCE ins,
  KHE_INDEXED_INSTANCE *ii)
{
  char *id;  KHE_INDEXED_INSTANCE ii2;  int i;
  id = KheInstanceId(ins);
  if( id != NULL )
  {
    if( HnTableRetrieve(archive->instance_table, id, ii2, i) )
      return *ii = ii2, true;
  }
  else
  {
    HaArrayForEach(archive->instance_array, ii2, i)
      if( ii2->instance == ins )
	return *ii = ii2, true;
  }
  return *ii = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveInstanceIndex(KHE_ARCHIVE archive, KHE_INSTANCE ins,      */
/*    int *index)                                                            */
/*                                                                           */
/*  Return the index of instance in archive.                                 */
/*                                                                           */
/*****************************************************************************/

bool KheArchiveContainsInstance(KHE_ARCHIVE archive, KHE_INSTANCE ins,
  int *index)
{
  KHE_INDEXED_INSTANCE ii;
  if( KheArchiveFindInstance(archive, ins, &ii) )
    return *index = ii->index, true;
  else
    return *index = -1, false;
}


/*****************************************************************************/
/*                                                                           */
/*  Sumbodule "solution groups"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveAddSolnGroup(KHE_ARCHIVE archive,                         */
/*    KHE_SOLN_GROUP soln_group)                                             */
/*                                                                           */
/*  Add soln_group to archive, assuming it is safe (not duplicate Id).       */
/*                                                                           */
/*****************************************************************************/

void KheArchiveAddSolnGroup(KHE_ARCHIVE archive, KHE_SOLN_GROUP soln_group)
{
  char *id;
  HaArrayAddLast(archive->soln_group_array, soln_group);
  id = KheSolnGroupId(soln_group);
  if( id != NULL )
    HnTableAdd(archive->soln_group_table, id, soln_group);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveDeleteSolnGroup(KHE_ARCHIVE archive,                      */
/*    KHE_SOLN_GROUP soln_group)                                             */
/*                                                                           */
/*  Delete soln_group from archive.  It must be present.                     */
/*                                                                           */
/*****************************************************************************/

void KheArchiveDeleteSolnGroup(KHE_ARCHIVE archive, KHE_SOLN_GROUP soln_group)
{
  char *id;  int pos;  KHE_SOLN_GROUP sg;
  if( !HaArrayContains(archive->soln_group_array, soln_group, &pos) )
    HnAbort("KheArchiveDeleteSolnGroup internal error 1");
  HaArrayDeleteAndShift(archive->soln_group_array, pos);
  id = KheSolnGroupId(soln_group);
  if( id != NULL )
  {
    if( !HnTableRetrieve(archive->soln_group_table, id, sg, pos) ||
	sg != soln_group )
      HnAbort("KheArchiveDeleteSolnGroup internal error 2");
    HnTableDelete(archive->soln_group_table, pos);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheArchiveSolnGroupCount(KHE_ARCHIVE archive)                        */
/*                                                                           */
/*  Return the number of solution groups in archive.                         */
/*                                                                           */
/*****************************************************************************/

int KheArchiveSolnGroupCount(KHE_ARCHIVE archive)
{
  return HaArrayCount(archive->soln_group_array);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_SOLN_GROUP KheArchiveSolnGroup(KHE_ARCHIVE archive, int i)           */
/*                                                                           */
/*  Return the i'th solution group of archive.                               */
/*                                                                           */
/*****************************************************************************/

KHE_SOLN_GROUP KheArchiveSolnGroup(KHE_ARCHIVE archive, int i)
{
  return HaArray(archive->soln_group_array, i);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveRetrieveSolnGroup(KHE_ARCHIVE archive, char *id,          */
/*    KHE_SOLN_GROUP *soln_group)                                            */
/*                                                                           */
/*  Retrieve a solution group with the given id from archive.                */
/*                                                                           */
/*****************************************************************************/

bool KheArchiveRetrieveSolnGroup(KHE_ARCHIVE archive, char *id,
  KHE_SOLN_GROUP *soln_group)
{
  int pos;
  return HnTableRetrieve(archive->soln_group_table, id, *soln_group, pos);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "reading and writing archives"                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveEltToArchive(KML_ELT archive_elt, KHE_ARCHIVE *archive,   */
/*    bool audit_and_fix, bool infer_resource_partitions,                    */
/*    bool allow_invalid_solns, KML_ERROR *ke, HA_ARENA a)                   */
/*                                                                           */
/*  Convert archive_elt into *archive.                                       */
/*                                                                           */
/*  Arena a is used only for error messages.                                 */
/*                                                                           */
/*****************************************************************************/

static bool KheArchiveEltToArchive(KML_ELT archive_elt, KHE_ARCHIVE *archive,
  bool audit_and_fix, bool infer_resource_partitions, /* bool with_monitors, */
  bool allow_invalid_solns, KML_ERROR *ke, HA_ARENA a)
{
  char *id;  KML_ELT metadata_elt, instances_elt, instance_elt;
  KML_ELT soln_groups_elt, soln_group_elt;  KHE_ARCHIVE res;  int i;
  KHE_INSTANCE ins;  KHE_MODEL model;

  /* fail if archive_elt has problems */
  *archive = NULL;
  if( strcmp(KmlLabel(archive_elt), "HighSchoolTimetableArchive") == 0 )
    model = KHE_MODEL_HIGH_SCHOOL_TIMETABLE;
  else if( strcmp(KmlLabel(archive_elt), "EmployeeScheduleArchive") == 0 )
    model = KHE_MODEL_EMPLOYEE_SCHEDULE;
  else
    return KmlError(ke, a, KmlLineNum(archive_elt), KmlColNum(archive_elt),
      "file does not begin with <HighSchoolTimetableArchive>"
      " or <EmployeeScheduleArchive>");
  if( !KmlCheck(archive_elt, "+Id : +MetaData +Instances +SolutionGroups", ke) )
    return false;

  /* create archive with optional id and optional metadata */
  id = KmlAttributeCount(archive_elt) == 0 ? NULL :
    KmlAttributeValue(archive_elt, 0);
  res = KheArchiveMake(id, NULL, model);
  if( KmlContainsChild(archive_elt, "MetaData", &metadata_elt) &&
      !KheArchiveMetaDataMakeFromKml(metadata_elt, res, ke) )
  {
    HnAssert(*ke != NULL, "KheArchiveEltToArchive internal error 1");
    return false;
  }

  /* build and add instances */
  if( KmlContainsChild(archive_elt, "Instances", &instances_elt) )
  {
    if( !KmlCheck(instances_elt, ": *Instance", ke) )
      return false;
    for( i = 0;  i < KmlChildCount(instances_elt);  i++ )
    {
      instance_elt = KmlChild(instances_elt, i);
      if( !KheInstanceMakeFromKml(instance_elt, model, audit_and_fix,
	    infer_resource_partitions, &ins, ke) )
      {
	HnAssert(*ke != NULL, "KheArchiveEltToArchive internal error 2");
	return false;
      }
      if( !KheArchiveAddInstance(res, ins) )
	return KmlError(ke, a,KmlLineNum(instance_elt), KmlColNum(instance_elt),
	  "instance Id \"%s\" used previously", KheInstanceId(ins));
    }
  }

  /* build and add solution groups */
  if( KmlContainsChild(archive_elt, "SolutionGroups", &soln_groups_elt) )
  {
    if( !KmlCheck(soln_groups_elt, ": *SolutionGroup", ke) )
      return false;
    for( i = 0;  i < KmlChildCount(soln_groups_elt);  i++ )
    {
      soln_group_elt = KmlChild(soln_groups_elt, i);
      if( !KheSolnGroupMakeFromKml(soln_group_elt, res,
	    /* with_monitors, */ allow_invalid_solns, ke) )
      {
	HnAssert(*ke != NULL, "KheArchiveEltToArchive internal error 3");
	return false;
      }
    }
  }
  *archive = res;
  *ke = NULL;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveRead(FILE *fp, KHE_ARCHIVE *archive, KML_ERROR *ke,       */
/*    bool audit_and_fix, bool infer_resource_partitions, bool with_monitors,*/
/*    bool allow_invalid_solns, char **leftover, int *leftover_len,          */
/*    FILE *echo_fp)                                                         */
/*                                                                           */
/*  Read *archive from fp.                                                   */
/*                                                                           */
/*****************************************************************************/

bool KheArchiveRead(FILE *fp, KHE_ARCHIVE *archive, KML_ERROR *ke,
  bool audit_and_fix, bool infer_resource_partitions, /* bool with_monitors, */
  bool allow_invalid_solns, FILE *echo_fp)
{
  return KheArchiveReadIncremental(fp, archive, ke, audit_and_fix,
    infer_resource_partitions, /* with_monitors, */ allow_invalid_solns,
    echo_fp, NULL, NULL, NULL, NULL, NULL, NULL);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveReadFromString(char *str, KHE_ARCHIVE *archive,           */
/*    KML_ERROR *ke, bool audit_and_fix, bool infer_resource_partitions,     */
/*    bool with_monitors, bool allow_invalid_solns)                          */
/*                                                                           */
/*  Like KheArchiveRead except that the archive is read from str.            */
/*                                                                           */
/*****************************************************************************/

bool KheArchiveReadFromString(char *str, KHE_ARCHIVE *archive,
  KML_ERROR *ke, bool audit_and_fix, bool infer_resource_partitions,
  /* bool with_monitors, */ bool allow_invalid_solns)
{
  KML_ELT archive_elt;  HA_ARENA a;
  a = HaArenaMake();
  return KmlReadString(str, &archive_elt, ke, a) &&
    KheArchiveEltToArchive(archive_elt, archive, audit_and_fix,
      infer_resource_partitions, /* with_monitors, */ allow_invalid_solns, ke,
      a);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "reading via the command line"                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void ArchiveExcludeInstances(KHE_ARCHIVE archive, char *ids)             */
/*                                                                           */
/*  Exclude instances with the given names from archive.                     */
/*                                                                           */
/*****************************************************************************/

static void ArchiveExcludeInstances(KHE_ARCHIVE archive, char *ids)
{
  char *id;  KHE_INSTANCE ins;  int junk;
  for( id = strtok(ids, ",");  id != NULL;  id = strtok(NULL, ",") )
    if( KheArchiveRetrieveInstance(archive, id, &ins, &junk) )
      KheArchiveDeleteInstance(archive, ins);
    else
      HnFatal("khe: no instance \"%s\" in archive", id);
}


/*****************************************************************************/
/*                                                                           */
/*  void ArchiveIncludeInstances(KHE_ARCHIVE archive, char *ids)             */
/*                                                                           */
/*  Include only instances with the given names from archive.                */
/*                                                                           */
/*****************************************************************************/

static void ArchiveIncludeInstances(KHE_ARCHIVE archive, char *ids)
{
  KHE_INSTANCE ins;  ARRAY_KHE_INSTANCE included_instances;  char *id;
  int i, junk;

  /* make sure all the named instances are present; build included_instances */
  HaArrayInit(included_instances, archive->arena);
  for( id = strtok(ids, ",");  id != NULL;  id = strtok(NULL, ",") )
    if( KheArchiveRetrieveInstance(archive, id, &ins, &junk) )
      HaArrayAddLast(included_instances, ins);
    else
      HnFatal("khe: no instance \"%s\" in archive", id);

  /* delete all instances not in included_instances */
  for( i = 0;  i < KheArchiveInstanceCount(archive);  i++ )
  {
    ins = KheArchiveInstance(archive, i);
    if( !HaArrayContains(included_instances, ins, &junk) )
    {
      KheArchiveDeleteInstance(archive, ins);
      i--;
    }
  }
  HaArrayFree(included_instances);
}


/*****************************************************************************/
/*                                                                           */
/*  void ArchiveExcludeSolutionGroups(KHE_ARCHIVE archive, char *ids)        */
/*                                                                           */
/*  Exclude solution groups with the given names from archive.               */
/*                                                                           */
/*****************************************************************************/

static void ArchiveExcludeSolutionGroups(KHE_ARCHIVE archive, char *ids)
{
  char *id;  KHE_SOLN_GROUP soln_group;
  if( ids[0] == '\0' )
  {
    /* special case, means delete all solution groups */
    while( KheArchiveSolnGroupCount(archive) > 0 )
      KheSolnGroupDelete(KheArchiveSolnGroup(archive, 0));
  }
  else
  {
    /* ordinary case, delete each solution group named by the option */
    for( id = strtok(ids, ",");  id != NULL;  id = strtok(NULL, ",") )
      if( KheArchiveRetrieveSolnGroup(archive, id, &soln_group) )
	KheSolnGroupDelete(soln_group);
      else
	HnFatal("khe: no solution group \"%s\" in archive", id);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void ArchiveIncludeSolutionGroups(KHE_ARCHIVE archive, char *ids)        */
/*                                                                           */
/*  Include only solution groups with the given names from archive.          */
/*                                                                           */
/*****************************************************************************/

static void ArchiveIncludeSolutionGroups(KHE_ARCHIVE archive, char *ids)
{
  KHE_SOLN_GROUP soln_group;  ARRAY_KHE_SOLN_GROUP included_soln_groups;
  char *id;  int i, junk;

  /* ensure all the named soln groups are present; build included_soln_groups */
  HaArrayInit(included_soln_groups, archive->arena);
  for( id = strtok(ids, ",");  id != NULL;  id = strtok(NULL, ",") )
    if( KheArchiveRetrieveSolnGroup(archive, id, &soln_group) )
      HaArrayAddLast(included_soln_groups, soln_group);
    else
      HnFatal("khe: no solution group \"%s\" in archive", id);

  /* delete all soln groups not in included_soln_groups */
  for( i = 0;  i < KheArchiveSolnGroupCount(archive);  i++ )
  {
    soln_group = KheArchiveSolnGroup(archive, i);
    if( !HaArrayContains(included_soln_groups, soln_group, &junk) )
    {
      KheSolnGroupDelete(soln_group);
      i--;
    }
  }
  HaArrayFree(included_soln_groups);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_ARCHIVE KheArchiveReadFromCommandLine(int argc, char *argv[],        */
/*    int *pos, bool audit_and_fix, bool infer_resource_partitions,          */
/*    bool allow_invalid_solns, FILE *echo_fp)                               */
/*                                                                           */
/*  Read an archive, returning only if successful.   The archive name is     */
/*  at argv[*pos], and following options cause it to be updated.             */
/*                                                                           */
/*****************************************************************************/

KHE_ARCHIVE KheArchiveReadFromCommandLine(int argc, char *argv[],
  int *pos, bool audit_and_fix, bool infer_resource_partitions,
  bool allow_invalid_solns, FILE *echo_fp)
{
  FILE *fp;  KHE_ARCHIVE res;  KML_ERROR ke;
  bool seen_x, seen_i, seen_X, seen_I;

  /* open the archive file for reading */
  if( *pos >= argc )
    HnFatal("khe: archive name missing from command line");
  if( strcmp(argv[*pos], "-") == 0 )
    fp = stdin;
  else
  {
    fp = fopen(argv[*pos], "r");
    if( fp == NULL )
      HnFatal("khe: cannot open file \"%s\" for reading", argv[*pos]);
  }

  /* read the archive file */
  if( !KheArchiveRead(fp, &res, &ke, audit_and_fix, infer_resource_partitions,
      allow_invalid_solns, echo_fp) )
    HnFatal("%s:%d:%d: %s\n", argv[*pos], KmlErrorLineNum(ke),
      KmlErrorColNum(ke), KmlErrorString(ke));

  /* process zero or more following -x, -i, -X, and -I options */
  seen_x = seen_i = seen_X = seen_I = false;
  for( (*pos)++;  *pos < argc && argv[*pos][0] == '-';  (*pos)++ )
    switch( argv[*pos][1] )
    {
      case 'x':

	if( seen_x )
	  HnFatal("khe: -x option used twice");
	if( seen_i )
	  HnFatal("khe: -x and -i used together");
	seen_x = true;
	ArchiveExcludeInstances(res, &argv[*pos][2]);
	break;

      case 'i':

	if( seen_i )
	  HnFatal("khe: -i option used twice");
	if( seen_x )
	  HnFatal("khe: -i and -x used together");
	seen_i = true;
	ArchiveIncludeInstances(res, &argv[*pos][2]);
	break;

      case 'X':

	if( seen_X )
	  HnFatal("khe: -X option used twice");
	if( seen_I )
	  HnFatal("khe: -X and -I used together");
	seen_X = true;
	ArchiveExcludeSolutionGroups(res, &argv[*pos][2]);
	break;

      case 'I':

	if( seen_I )
          HnFatal("khe: -I option used twice");
	if( seen_X )
          HnFatal("khe: -I and -X used together");
	seen_I = true;
	ArchiveIncludeSolutionGroups(res, &argv[*pos][2]);
	break;

      default:

	/* stop at first unrecognized option */
	return res;
    }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "incremental reading"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KmlEltHasRootLabel(KML_ELT elt)                                     */
/*                                                                           */
/*  Return true if elt has the appropriate label for a root element:         */
/*  <HighSchoolTimetableArchive> or <EmployeeScheduleArchive>.               */
/*                                                                           */
/*****************************************************************************/

bool KmlEltHasRootLabel(KML_ELT elt)
{
  return strcmp(KmlLabel(elt), "HighSchoolTimetableArchive") == 0 ||
    strcmp(KmlLabel(elt), "EmployeeScheduleArchive") == 0;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheCheckSolnGroupContext(KML_ELT sg, KML_READ_INFO ri)              */
/*                                                                           */
/*  Check that sg, whose label is <SolutionGroup>, has occurred in the       */
/*  correct context.                                                         */
/*                                                                           */
/*****************************************************************************/

static void KheCheckSolnGroupContext(KML_ELT sg, KML_READ_INFO ri)
{
  KML_ELT sgs, hsta;  HA_ARENA a;
  sgs = KmlParent(sg);
  a = KmlReadArena(ri);
  if( sgs == NULL || strcmp(KmlLabel(sgs), "SolutionGroups") != 0 )
    KmlReadFail(ri, KmlErrorMake(a, KmlLineNum(sg), KmlColNum(sg),
      "<SolutionGroup> does not lie within <SolutionGroups>"));
  hsta = KmlParent(sgs);
  if( hsta == NULL || !KmlEltHasRootLabel(hsta) )
    KmlReadFail(ri, KmlErrorMake(a, KmlLineNum(sgs), KmlColNum(sgs),
      "<SolutionGroups> does not lie within <HighSchoolTimetableArchive>"
      " or <EmployeeScheduleArchive>"));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheCheckSolnContext(KML_ELT s, KML_READ_INFO ri)                    */
/*                                                                           */
/*  Check that s, whose label is <Solution>, has occurred in the correct     */
/*  context.                                                                 */
/*                                                                           */
/*****************************************************************************/

static void KheCheckSolnContext(KML_ELT s, KML_READ_INFO ri)
{
  KML_ELT sg;
  sg = KmlParent(s);
  if( sg == NULL || strcmp(KmlLabel(sg), "SolutionGroup") != 0 )
    KmlReadFail(ri, KmlErrorMake(KmlReadArena(ri), KmlLineNum(s), KmlColNum(s),
      "<Solution> does not lie within <SolutionGroup>"));
  KheCheckSolnGroupContext(sg, ri);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheCheckInstanceContext(KML_ELT ins, KML_READ_INFO ri)              */
/*                                                                           */
/*  Check that ins, whose label is <Instance>, has occurred in the correct   */
/*  context.                                                                 */
/*                                                                           */
/*****************************************************************************/

static void KheCheckInstanceContext(KML_ELT ins, KML_READ_INFO ri)
{
  KML_ELT inss, hsta;
  inss = KmlParent(ins);
  if( inss == NULL || strcmp(KmlLabel(inss), "Instances") != 0 )
    KmlReadFail(ri, KmlErrorMake(KmlReadArena(ri), KmlLineNum(ins),
      KmlColNum(ins), "<Instance> does not lie within <Instances>"));
  hsta = KmlParent(inss);
  if( hsta == NULL || !KmlEltHasRootLabel(hsta) )
    KmlReadFail(ri, KmlErrorMake(KmlReadArena(ri), KmlLineNum(inss),
      KmlColNum(inss), "<Instances> does not lie within "
      "<HighSchoolTimetableArchive> or <EmployeeScheduleArchive> "
      "(it lies within %s)", hsta == NULL ? "nothing" : KmlLabel(hsta)));
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_READ_STATE - info to pass around during incremental file reads       */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_read_state_rec {
  KHE_ARCHIVE		archive;
  KHE_SOLN_GROUP	soln_group;
  bool			audit_and_fix;
  bool			infer_resource_partitions;
  /* bool		with_monitors; */
  bool			allow_invalid_solns;
  KHE_ARCHIVE_FN	archive_begin_fn;
  KHE_ARCHIVE_FN	archive_end_fn;
  KHE_SOLN_GROUP_FN	soln_group_begin_fn;
  KHE_SOLN_GROUP_FN	soln_group_end_fn;
  KHE_SOLN_FN		soln_fn;
  void			*impl;
} *KHE_READ_STATE;


/*****************************************************************************/
/*                                                                           */
/*  void KheReadBeginArchive(KHE_READ_STATE krs, KHE_ARCHIVE archive)        */
/*                                                                           */
/*  Function to call when beginning an archive.                              */
/*                                                                           */
/*****************************************************************************/

static void KheReadBeginArchive(KHE_READ_STATE krs, KHE_ARCHIVE archive)
{
  HnAssert(krs->archive == NULL, "KheReadBeginArchive internal error 1");
  krs->archive = archive;
  if( krs->archive_begin_fn != NULL )
    krs->archive_begin_fn(krs->archive, krs->impl);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheReadEndArchive(KHE_READ_STATE krs)                               */
/*                                                                           */
/*  Function to call when ending an archive.                                 */
/*                                                                           */
/*****************************************************************************/

static void KheReadEndArchive(KHE_READ_STATE krs)
{
  HnAssert(krs->archive != NULL, "KheReadEndArchive internal error 1");
  if( krs->archive_end_fn != NULL )
    krs->archive_end_fn(krs->archive, krs->impl);
  krs->archive = NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheReadInArchive(KHE_READ_STATE krs)                                */
/*                                                                           */
/*  Return true if krs is currently within an archive.                       */
/*                                                                           */
/*****************************************************************************/

static bool KheReadInArchive(KHE_READ_STATE krs)
{
  return krs->archive != NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheReadBeginSolnGroup(KHE_READ_STATE krs, KHE_SOLN_GROUP soln_group)*/
/*                                                                           */
/*  Function to call when beginning a soln group.                            */
/*                                                                           */
/*****************************************************************************/

static void KheReadBeginSolnGroup(KHE_READ_STATE krs, KHE_SOLN_GROUP soln_group)
{
  HnAssert(krs->archive != NULL, "KheReadBeginSolnGroup internal error 1");
  HnAssert(krs->soln_group == NULL, "KheReadBeginSolnGroup internal error 2");
  krs->soln_group = soln_group;
  if( krs->soln_group_begin_fn != NULL )
    krs->soln_group_begin_fn(krs->soln_group, krs->impl);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheReadEndSolnGroup(KHE_READ_STATE krs, KHE_SOLN_GROUP soln_group)  */
/*                                                                           */
/*  Function to call when ending a soln group.                               */
/*                                                                           */
/*****************************************************************************/

static void KheReadEndSolnGroup(KHE_READ_STATE krs, KHE_SOLN_GROUP soln_group)
{
  HnAssert(krs->archive != NULL, "KheReadEndSolnGroup internal error 1");
  HnAssert(krs->soln_group != NULL, "KheReadEndSolnGroup internal error 2");
  HnAssert(krs->soln_group==soln_group,"KheReadEndSolnGroup internal error 3");
  if( krs->soln_group_end_fn != NULL )
    krs->soln_group_end_fn(krs->soln_group, krs->impl);
  krs->soln_group = NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheReadInSolnGroup(KHE_READ_STATE krs)                              */
/*                                                                           */
/*  Return true if krs is currently within a solution group.                 */
/*                                                                           */
/*****************************************************************************/

static bool KheReadInSolnGroup(KHE_READ_STATE krs)
{
  return krs->soln_group != NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheReadEltFn(KML_ELT elt, KML_READ_INFO ri)                         */
/*                                                                           */
/*  Element function called during incremental reading.                      */
/*                                                                           */
/*****************************************************************************/

static void KheReadEltFn(KML_ELT elt, KML_READ_INFO ri)
{
  KML_ERROR ke;  int count;  KHE_SOLN soln;  KHE_INSTANCE ins;
  KHE_ARCHIVE archive;  KHE_SOLN_GROUP soln_group;
  KHE_READ_STATE krs = (KHE_READ_STATE) KmlReadImpl(ri);
  if( DEBUG2 )
    fprintf(stderr, "[ KheReadEltFn(%s, %d)\n",
      KmlLabel(elt), KmlReadCurrDepth(ri));

  /* after completing the read of an instance */
  ke = NULL;
  if( strcmp(KmlLabel(elt), "Instance") == 0 )
  {
    KheCheckInstanceContext(elt, ri);
    if( !KheReadInArchive(krs) )
    {
      /* this must be the very first instance */
      if( !KheArchiveEltToArchive(KmlParent(KmlParent(elt)), &archive,
	    krs->audit_and_fix, krs->infer_resource_partitions,
	    /* krs->with_monitors, */ krs->allow_invalid_solns, &ke,
	    KmlReadArena(ri)) )
      {
	HnAssert(ke != NULL, "KheReadEltFn internal error 1");
	KmlReadFail(ri, ke);
      }
      KheReadBeginArchive(krs, archive);
      count = KheArchiveSolnGroupCount(krs->archive);
      HnAssert(count == 0, "KheReadEltFn internal error 2");
      count = KheArchiveInstanceCount(krs->archive);
      HnAssert(count == 1, "KheReadEltFn internal error 3");
      if( DEBUG2 )
      {
	ins = KheArchiveInstance(krs->archive, 0);
	fprintf(stderr, "  adding initial instance %s to archive\n",
	  KheInstanceId(ins));
      }
    }
    else
    {
      /* this must be at least the second instance in this archive */
      if( !KheInstanceMakeFromKml(elt, KheArchiveModel(krs->archive),
	    krs->audit_and_fix, krs->infer_resource_partitions, &ins, &ke) )
      {
	HnAssert(ke != NULL, "KheReadEltFn internal error 4");
        KmlReadFail(ri, ke);
      }
      if( !KheArchiveAddInstance(krs->archive, ins) )
      {
	KmlError(&ke, KmlReadArena(ri), KmlLineNum(elt), KmlColNum(elt),
	  "instance Id \"%s\" used previously", KheInstanceId(ins));
	KmlReadFail(ri, ke);
      }
      count = KheArchiveInstanceCount(krs->archive);
      HnAssert(count >= 2, "KheReadEltFn internal error 5");
      if( DEBUG2 )
      {
	ins = KheArchiveInstance(krs->archive, count - 1);
	fprintf(stderr, "  adding instance %s to archive\n",
	  KheInstanceId(ins));
      }
    }
    KmlDeleteChild(KmlParent(elt), elt);
    KmlFree(elt);  /* still to do here */
  }

  /* after completing the read of a solution */
  else if( strcmp(KmlLabel(elt), "Solution") == 0 )
  {
    KheCheckSolnContext(elt, ri);
    if( !KheReadInArchive(krs) )
    {
      /* this must be the very first solution */
      if( !KheArchiveEltToArchive(KmlParent(KmlParent(KmlParent(elt))),
	    &archive, krs->audit_and_fix, krs->infer_resource_partitions,
	    /* krs->with_monitors, */ krs->allow_invalid_solns, &ke,
	    KmlReadArena(ri)) )
      {
	HnAssert(ke != NULL, "KheReadEltFn internal error 6");
	KmlReadFail(ri, ke);
      }
      KheReadBeginArchive(krs, archive);
      count = KheArchiveSolnGroupCount(krs->archive);
      HnAssert(count == 1, "KheReadEltFn internal error 7");
      soln_group = KheArchiveSolnGroup(krs->archive, 0);
      HnAssert(KheSolnGroupSolnCount(soln_group) == 1,
	"KheReadEltFn internal error 2");
      KheReadBeginSolnGroup(krs, soln_group);
      if( DEBUG2 )
      {
	soln = KheSolnGroupSoln(krs->soln_group, 0);
	fprintf(stderr, "  adding initial solution (to %s) to soln group %s\n",
	  KheInstanceId(KheSolnInstance(soln)),KheSolnGroupId(krs->soln_group));
      }
    }
    else if( !KheReadInSolnGroup(krs) )
    {
      /* this must be the first solution in this solution group */
      if( !KheSolnGroupMakeFromKml(KmlParent(elt), krs->archive,
	    /* krs->with_monitors, */ krs->allow_invalid_solns, &ke) )
      {
	HnAssert(ke != NULL, "KheReadEltFn internal error 8");
	KmlReadFail(ri, ke);
      }
      count = KheArchiveSolnGroupCount(krs->archive);
      HnAssert(count >= 1, "KheReadEltFn internal error 9");
      soln_group = KheArchiveSolnGroup(krs->archive, count - 1);
      HnAssert(KheSolnGroupSolnCount(soln_group) == 1,
	"KheReadEltFn internal error 10");
      KheReadBeginSolnGroup(krs, soln_group);
      if( DEBUG2 )
      {
	soln = KheSolnGroupSoln(krs->soln_group, 0);
	fprintf(stderr, "  adding first solution (to %s) to soln group %s\n",
	  KheInstanceId(KheSolnInstance(soln)),KheSolnGroupId(krs->soln_group));
      }
    }
    else
    {
      /* this must be at least the second solution in this solution group */
      if( !KheSolnMakeFromKml(elt, krs->archive, /* krs->with_monitors, */
	    krs->allow_invalid_solns, &soln, &ke) )
      {
	HnAssert(ke != NULL, "KheReadEltFn internal error 11");
        KmlReadFail(ri, ke);
      }
      KheSolnGroupAddSoln(krs->soln_group, soln);
      if( DEBUG2 )
	fprintf(stderr, "  adding solution to soln group %s\n",
          KheSolnGroupId(krs->soln_group));
    }
    KmlDeleteChild(KmlParent(elt), elt);
    KmlFree(elt);
    if( krs->soln_fn != NULL )
    {
      count = KheSolnGroupSolnCount(krs->soln_group);
      HnAssert(count >= 1, "KheReadEltFn internal error 11a");
      soln = KheSolnGroupSoln(krs->soln_group, count - 1);
      if( DEBUG2 && krs->soln_fn != NULL )
	fprintf(stderr, "  calling soln_fn on soln to %s\n",
	  KheInstanceId(KheSolnInstance(soln)));
      krs->soln_fn(soln, krs->impl);
    }
  }

  /* after completing the read of a solution group */
  else if( strcmp(KmlLabel(elt), "SolutionGroup") == 0 )
  {
    KheCheckSolnGroupContext(elt, ri);
    if( !KheReadInArchive(krs) )
    {
      /* this must be the very first solution group, and it must contain */
      /* no solutions, which is unlikely but legal */
      if( !KheArchiveEltToArchive(KmlParent(KmlParent(elt)), &archive,
	    krs->audit_and_fix, krs->infer_resource_partitions,
	    /* krs->with_monitors, */ krs->allow_invalid_solns, &ke,
	    KmlReadArena(ri)) )
      {
	HnAssert(ke != NULL, "KheReadEltFn internal error 12");
	KmlReadFail(ri, ke);
      }
      KheReadBeginArchive(krs, archive);
      count = KheArchiveSolnGroupCount(krs->archive);
      HnAssert(count == 1, "KheReadEltFn internal error 13");
      soln_group = KheArchiveSolnGroup(krs->archive, 0);
      KheReadBeginSolnGroup(krs, soln_group);
      KheReadEndSolnGroup(krs, soln_group);
      if( DEBUG2 )
	fprintf(stderr, "  adding empty first solution group to archive\n");
    }
    else if( !KheReadInSolnGroup(krs) )
    {
      /* this must be a solution group after the first, and it must */
      /* contain no solutions */
      if( !KheSolnGroupMakeFromKml(elt, krs->archive,
	    /* krs->with_monitors, */ krs->allow_invalid_solns, &ke) )
      {
	HnAssert(ke != NULL, "KheReadEltFn internal error 14");
	KmlReadFail(ri, ke);
      }
      count = KheArchiveSolnGroupCount(krs->archive);
      soln_group = KheArchiveSolnGroup(krs->archive, count - 1);
      KheReadBeginSolnGroup(krs, soln_group);
      KheReadEndSolnGroup(krs, soln_group);
      if( DEBUG2 )
	fprintf(stderr, "  adding empty solution group to archive\n");
    }
    else
    {
      /* must have seen at least one solution, which will have done */
      /* all the real work.  So just end this solution group now    */
      if( DEBUG2 )
	fprintf(stderr, "  ending soln group %s\n",
          KheSolnGroupId(krs->soln_group));
      KheReadEndSolnGroup(krs, krs->soln_group);
    }
  }
  if( DEBUG2 )
    fprintf(stderr, "] KheReadEltFn returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveReadIncremental(FILE *fp, KHE_ARCHIVE *archive,           */
/*    KML_ERROR *ke, bool audit_and_fix, bool infer_resource_partitions,     */
/*    bool with_monitors, bool allow_invalid_solns, FILE *echo_fp,           */
/*    KHE_ARCHIVE_FN archive_begin_fn, KHE_ARCHIVE_FN archive_end_fn,        */
/*    KHE_SOLN_GROUP_FN soln_group_begin_fn,                                 */
/*    KHE_SOLN_GROUP_FN soln_group_end_fn, KHE_SOLN_FN soln_fn, void *impl)  */
/*                                                                           */
/*  Like KheArchiveRead but also calls the callback functions as needed.     */
/*                                                                           */
/*  Implementation note.  A callback to KheReadEltFn above occurs as each    */
/*  category of depth at most 3 is completed.  Here is the grammar:          */
/*                                                                           */
/*    HighSchoolTimetableArchive +Id                                         */
/*       +MetaData                                                           */
/*       +Instances                                                          */
/*          *Instance                                                        */
/*       +SolutionGroups                                                     */
/*          *SolutionGroup Id                                                */
/*             MetaData                                                      */
/*             *Solution                                                     */
/*                                                                           */
/*  HighSchoolTimetableArchive may be replaced by EmployeeScheduleArchive.   */
/*                                                                           */
/*****************************************************************************/

bool KheArchiveReadIncremental(FILE *fp, KHE_ARCHIVE *archive,
  KML_ERROR *ke, bool audit_and_fix, bool infer_resource_partitions,
  /* bool with_monitors, */ bool allow_invalid_solns, FILE *echo_fp,
  KHE_ARCHIVE_FN archive_begin_fn, KHE_ARCHIVE_FN archive_end_fn,
  KHE_SOLN_GROUP_FN soln_group_begin_fn,
  KHE_SOLN_GROUP_FN soln_group_end_fn, KHE_SOLN_FN soln_fn, void *impl)
{
  struct khe_read_state_rec krs_rec;  KML_ELT archive_elt;  bool res;
  HA_ARENA a;

  /* initialize archive and incremental read data */
  a = HaArenaMake();
  krs_rec.archive = NULL;
  krs_rec.soln_group = NULL;
  krs_rec.audit_and_fix = audit_and_fix;
  krs_rec.infer_resource_partitions = infer_resource_partitions;
  /* krs_rec.with_monitors = with_monitors; */
  krs_rec.allow_invalid_solns = allow_invalid_solns;
  krs_rec.archive_begin_fn = archive_begin_fn;
  krs_rec.archive_end_fn = archive_end_fn;
  krs_rec.soln_group_begin_fn = soln_group_begin_fn;
  krs_rec.soln_group_end_fn = soln_group_end_fn;
  krs_rec.soln_fn = soln_fn;
  krs_rec.impl = impl;

  /* read fp incrementally, calling back after each high-level thing */
  *ke = NULL;
  if( !KmlReadFileIncremental(fp, &archive_elt, ke, echo_fp, &KheReadEltFn,
      &krs_rec, 3, a) )
  {
    /* the incremental read did not work */
    *archive = NULL;
    HnAssert(*ke != NULL, "KheArchiveReadIncremental internal error 1");
    if( DEBUG3 )
      fprintf(stderr, "KheArchiveReadIncremental ret false (a)\n");
    res = false;
  }
  else if( KheReadInSolnGroup(&krs_rec) )
  {
    /* the incremental read ended up inside a solution group */
    *archive = NULL;
    HnAssert(*ke != NULL, "KheArchiveReadIncremental internal error 2");
    if( DEBUG3 )
      fprintf(stderr, "KheArchiveReadIncremental ret false (b)\n");
    res = false;
  }
  else if( KheReadInArchive(&krs_rec) )
  {
    /* the incremental read worked and found at least one soln or soln group */
    *archive = krs_rec.archive;
    KheReadEndArchive(&krs_rec);
    res = true;
    HnAssert(KmlParent(archive_elt) == NULL,
      "KheArchiveReadIncremental internal error 3");
    KmlFree(archive_elt);
    if( DEBUG3 )
      fprintf(stderr, "KheArchiveReadIncremental ret true (c)\n");
  }
  else
  {
    /* the incremental read worked but found no solns or soln groups */
    res = KheArchiveEltToArchive(archive_elt, archive, audit_and_fix,
      infer_resource_partitions, /* with_monitors, */ allow_invalid_solns,
      ke, a);
    KheReadBeginArchive(&krs_rec, *archive);
    KheReadEndArchive(&krs_rec);
    HnAssert(KmlParent(archive_elt) == NULL,
      "KheArchiveReadIncremental internal error 4");
    KmlFree(archive_elt);
    if( !res )
    {
      HnAssert(*ke != NULL, "KheArchiveReadIncremental internal error 5");
      if( DEBUG3 )
	fprintf(stderr, "KheArchiveReadIncremental ret false (d)\n");
    }
    else if( DEBUG3 )
      fprintf(stderr, "KheArchiveReadIncremental ret true (e)\n");
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheArchiveInitialTag(KHE_ARCHIVE archive)                          */
/*                                                                           */
/*  Return the initial tag needed for a file containing archive.             */
/*                                                                           */
/*****************************************************************************/

static char *KheArchiveInitialTag(KHE_ARCHIVE archive)
{
  switch( KheArchiveModel(archive) )
  {
    case KHE_MODEL_HIGH_SCHOOL_TIMETABLE: return "HighSchoolTimetableArchive";
    case KHE_MODEL_EMPLOYEE_SCHEDULE:	  return "EmployeeScheduleArchive";

    default:

      HnAbort("KheArchiveInitialTag internal error");
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheArchiveWriteHeaderAndInstances(KHE_ARCHIVE archive, KML_FILE kf) */
/*                                                                           */
/*  Write the header and instances of archive to kf.                         */
/*                                                                           */
/*****************************************************************************/

static void KheArchiveWriteHeaderAndInstances(KHE_ARCHIVE archive, KML_FILE kf)
{
  int i;  KHE_INDEXED_INSTANCE ii;

  /* header with optional Id, followed by optional metadata */
  KmlBegin(kf, KheArchiveInitialTag(archive));
  if( archive->id != NULL )
    KmlAttribute(kf, "Id", archive->id);
  if( archive->meta_data != NULL )
    KheArchiveMetaDataWrite(archive->meta_data, kf);

  /* instances */
  if( HaArrayCount(archive->instance_array) > 0 )
  {
    KmlBegin(kf, "Instances");
    HaArrayForEach(archive->instance_array, ii, i)
      KheInstanceWrite(ii->instance, kf);
    KmlEnd(kf, "Instances");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveWrite(KHE_ARCHIVE archive, bool with_reports, FILE *fp)   */
/*                                                                           */
/*  Write archive to fp, which must be open for writing.  Include reports    */
/*  in the written archive if with_reports is true.                          */
/*                                                                           */
/*****************************************************************************/

void KheArchiveWrite(KHE_ARCHIVE archive, bool with_reports, FILE *fp)
{
  KML_FILE kf;  int i;  KHE_SOLN_GROUP soln_group;

  /* header and instances */
  kf = KmlMakeFile(fp, 0, 2);
  KheArchiveWriteHeaderAndInstances(archive, kf);

  /* soln groups */
  if( HaArrayCount(archive->soln_group_array) > 0 )
  {
    KmlBegin(kf, "SolutionGroups");
    HaArrayForEach(archive->soln_group_array, soln_group, i)
      KheSolnGroupWrite(soln_group, with_reports, kf);
    KmlEnd(kf, "SolutionGroups");
  }

  /* close header and exit */
  KmlEnd(kf, KheArchiveInitialTag(archive));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveWriteSolnGroup(KHE_ARCHIVE archive,                       */
/*    KHE_SOLN_GROUP soln_group, bool with_reports, FILE *fp)                */
/*                                                                           */
/*  Write archive, but only include one of its solution groups:  soln_group. */
/*                                                                           */
/*****************************************************************************/

void KheArchiveWriteSolnGroup(KHE_ARCHIVE archive,
  KHE_SOLN_GROUP soln_group, bool with_reports, FILE *fp)
{
  KML_FILE kf;

  /* header and instances */
  HnAssert(KheSolnGroupArchive(soln_group) == archive,
    "KheArchiveWriteSolnGroup:  soln_group not in archive");
  kf = KmlMakeFile(fp, 0, 2);
  KheArchiveWriteHeaderAndInstances(archive, kf);

  /* soln group */
  KmlBegin(kf, "SolutionGroups");
  KheSolnGroupWrite(soln_group, with_reports, kf);
  KmlEnd(kf, "SolutionGroups");

  /* close header and exit */
  KmlEnd(kf, KheArchiveInitialTag(archive));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveWriteWithoutSolnGroups(KHE_ARCHIVE archive, FILE *fp)     */
/*                                                                           */
/*  Write archive, omitting its solution groups.                             */
/*                                                                           */
/*****************************************************************************/

void KheArchiveWriteWithoutSolnGroups(KHE_ARCHIVE archive, FILE *fp)
{
  KML_FILE kf;

  /* header and instances */
  kf = KmlMakeFile(fp, 0, 2);
  KheArchiveWriteHeaderAndInstances(archive, kf);

  /* close header and exit */
  KmlEnd(kf, KheArchiveInitialTag(archive));
}
