
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sm_parallel_solve.c                                    */
/*  DESCRIPTION:  KheParallelSolve().  If you can't compile this file,       */
/*                see the makefile for a workaround.                         */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "howard_a.h"
#include "howard_n.h"
#include <limits.h>
#include <math.h>
#if KHE_USE_PTHREAD
#include <pthread.h>
#endif

#define DEBUG1 0
#define DEBUG2 1


/*****************************************************************************/
/*                                                                           */
/*  PINST - an instance to be solved in parallel                             */
/*                                                                           */
/*  Keeping track of where we are up to with an instance is quite tricky.    */
/*  We use three variables:                                                  */
/*                                                                           */
/*    wanted_solves                                                          */
/*      The number of solves of pinst->instance that we want to do.          */
/*      Initially this will be psolver->ps_make, but if time runs out        */
/*      it is reduced to initiated_solves.                                   */
/*                                                                           */
/*    initiated_solves                                                       */
/*      The number of solves of pinst->instance that have been initiated.    */
/*                                                                           */
/*  The number of completed solves is not really needed, although it is      */
/*  available as HaArrayCount(pinst->all_costs).                             */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_SOLN) ARRAY_KHE_SOLN;
typedef HA_ARRAY(KHE_COST) ARRAY_KHE_COST;

typedef struct pinst_rec {
  /* PSOLVER		psolver; */
  KHE_INSTANCE		instance;		/* the instance to solve     */
  KHE_STATS_TIMER	ins_timer;		/* the time the solve began  */
  float			running_time;		/* the final running time    */
  float			ps_time_limit;		/* time limit (-1.0 if none) */
  ARRAY_KHE_SOLN	best_solns;		/* best solns so far         */
  int			ps_keep;		/* keep this many solns      */
  int			wanted_solves;		/* wanted solves (ps_make)   */
  int			initiated_solves;	/* initiated solves          */
  ARRAY_KHE_COST	all_costs;		/* costs of all solutions    */
  ARRAY_KHE_COST	unique_costs;		/* uniqueified all_costs     */
} *PINST;

typedef HA_ARRAY(PINST) ARRAY_PINST;


/*****************************************************************************/
/*                                                                           */
/*  PSOLVER and PTHREAD                                                      */
/*                                                                           */
/*****************************************************************************/

typedef struct psolver_rec *PSOLVER;
typedef struct pthread_rec *PTHREAD;


/*****************************************************************************/
/*                                                                           */
/*  PTHREAD - one thread                                                     */
/*                                                                           */
/*****************************************************************************/

#if KHE_USE_PTHREAD
#define PTHREAD_FIELD pthread_t thread;
#else
#define PTHREAD_FIELD
#endif

#define INHERIT_PTHREAD							\
  HA_ARENA		arena;						\
  PSOLVER		psolver;					\
  KHE_OPTIONS		options;					\
  HA_ARENA_SET		arena_set;					\
  PTHREAD_FIELD

struct pthread_rec {
  INHERIT_PTHREAD
};

typedef HA_ARRAY(PTHREAD) ARRAY_PTHREAD;


/*****************************************************************************/
/*                                                                           */
/*  PSOLVER - a parallel solver                                              */
/*                                                                           */
/*  Implementation note.  PSOLVER is a subtype of PTHREAD; as well as being  */
/*  a solver object, it represents the leader thread, the thread that is     */
/*  already running to begin with.                                           */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_PS_TIME_OMIT,
  KHE_PS_TIME_SHARED
} KHE_PS_TIME;

struct psolver_rec {
  INHERIT_PTHREAD
  KHE_GENERAL_SOLVER	solver;			/* solver parameter          */
  KHE_STATS_TIMER	global_timer;		/* times whole run           */
  int			ps_make;		/* ps_make option            */
  bool			ps_no_diversify;	/* ps_no_diversify option    */
  int			ps_keep;		/* ps_keep option            */
  KHE_PS_TIME		ps_time_measure;	/* ps_time_measure option    */
  float			ps_time_limit;		/* time limit, or -1 if none */
  int			curr_instance;		/* currently on this         */
  ARRAY_PINST		instances;		/* instances to be solved    */
#if KHE_USE_PTHREAD
  int			ps_threads;		/* ps_threads option         */
  ARRAY_PTHREAD		other_threads;		/* non-leader threads        */
  pthread_mutex_t	mutex;			/* locks everything          */
#endif
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "PInst"                                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  PINST PInstMake(KHE_INSTANCE ins, int wanted_solves,                     */
/*    float ps_time_limit, int ps_keep, HA_ARENA a)                          */
/*                                                                           */
/*  Make a new pinst object, for solving ins.                                */
/*                                                                           */
/*****************************************************************************/

static PINST PInstMake(KHE_INSTANCE ins, int wanted_solves,
  float ps_time_limit, int ps_keep, HA_ARENA a)
{
  PINST res;
  HaMake(res, a);
  /* res->psolver = psolver; */
  res->instance = ins;
  res->ins_timer = KheStatsTimerMake(a);
  res->running_time = 0.0;
  res->ps_time_limit = ps_time_limit;
  HaArrayInit(res->best_solns, a);
  res->ps_keep = ps_keep;
  res->wanted_solves = wanted_solves;
  res->initiated_solves = 0;
  HaArrayInit(res->all_costs, a);
  HaArrayInit(res->unique_costs, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool PInstTimeLimitNotReached(PINST pinst)                               */
/*                                                                           */
/*  Return true if pinst's time limit has not yet been reached.  A limit     */
/*  of -1.0 means no limit.                                                  */
/*                                                                           */
/*****************************************************************************/

static bool PInstTimeLimitNotReached(PINST pinst)
{
  return pinst->ps_time_limit == -1.0 ||
    KheStatsTimerNow(pinst->ins_timer) < pinst->ps_time_limit;
}


/*****************************************************************************/
/*                                                                           */
/*  bool PInstSolveWanted(PINST pinst, int *diversifier)                     */
/*                                                                           */
/*  If pinst wants a solve, that is, if it has not yet initiated ps_make     */
/*  solves and time is available, return true, otherwise false.              */
/*                                                                           */
/*  If true is returned, pinst understands that a new solve will actually    */
/*  be initiated, and it updates itself accordingly, and returns the         */
/*  appropriate diversifier to use for the new solve.  This diversifier      */
/*  will be ignored if the ps_no_diversify option is set.                    */
/*                                                                           */
/*  This function must be run while the enclosing psolver is locked.         */
/*  Accordingly we do as little as possible; in particular we do not         */
/*  create a new solution.                                                   */
/*                                                                           */
/*****************************************************************************/

static bool PInstSolveWanted(PINST pinst, int *diversifier)
{
  if( pinst->initiated_solves == 0 )
    KheStatsTimerReset(pinst->ins_timer);
  if( pinst->initiated_solves < pinst->wanted_solves &&
      PInstTimeLimitNotReached(pinst) )
  {
    /* OK to start off another solve of pinst */
    *diversifier = pinst->initiated_solves++;
    if( DEBUG2 )
      fprintf(stderr, "  parallel solve of %s: starting solve %d%s\n",
	KheInstanceId(pinst->instance), pinst->initiated_solves,
	pinst->initiated_solves == pinst->wanted_solves ? " (last)" : "");
    return true;
  }
  else
  {
    /* no more solves of pinst */
    if( DEBUG2 )
      if( pinst->initiated_solves < pinst->wanted_solves )
	fprintf(stderr, "  parallel solve of %s: only %d of %d solves "
	  "started (time limit reached)\n", KheInstanceId(pinst->instance),
	  pinst->initiated_solves, pinst->wanted_solves);
    pinst->wanted_solves = pinst->initiated_solves;
    *diversifier = -1;
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void PInstSolveEnded(PINST pinst, KHE_SOLN soln, KHE_SOLN *unwanted_soln)*/
/*                                                                           */
/*  Add soln to pinst.  If *unwanted_soln is non-NULL on return, it is not   */
/*  wanted and should be deleted.                                            */
/*                                                                           */
/*  This function must be run while the enclosing psolver is locked.         */
/*  Accordingly we do as little as possible; in particular we do not         */
/*  delete any unwanted solution.                                            */
/*                                                                           */
/*****************************************************************************/
static void PInstDebug(PINST pinst, int indent, FILE *fp);

static void PInstSolveEnded(PINST pinst, KHE_SOLN soln, KHE_SOLN *unwanted_soln)
{
  int i;  KHE_SOLN best_soln;
  HaArrayAddLast(pinst->all_costs, KheSolnCost(soln));
  HaArrayAddLast(pinst->unique_costs, KheSolnCost(soln));
  for( i = 0;  i < HaArrayCount(pinst->best_solns);  i++ )
  {
    best_soln = HaArray(pinst->best_solns, i);
    if( KheSolnCost(soln) < KheSolnCost(best_soln) )
      break;
  }
  HaArrayAdd(pinst->best_solns, i, soln);  /* works even at and */
  if( HaArrayCount(pinst->best_solns) > pinst->ps_keep )
    *unwanted_soln = HaArrayLastAndDelete(pinst->best_solns);
  else
    *unwanted_soln = NULL;
  pinst->running_time = KheStatsTimerNow(pinst->ins_timer);
  if( DEBUG2 && HaArrayCount(pinst->all_costs) == pinst->wanted_solves )
  {
    fprintf(stderr, "  parallel solve of %s:  ending last of %d solves\n",
      KheInstanceId(pinst->instance), pinst->wanted_solves);
    PInstDebug(pinst, 4, stderr);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void PInstEndRun(PINST pinst, KHE_SOLN_GROUP soln_group,                 */
/*    HA_ARENA_SET final_arena_set, bool set_time)                           */
/*                                                                           */
/*  Solving of pinst has ended.  Don't actually delete pinst (it gets        */
/*  deleted when the solver's arena is deleted), but do what is needed       */
/*  when pinst will not be used any more.  If soln_group is NULL, delete     */
/*  all the saved solutions.  Otherwise, reset their arena sets and          */
/*  optionally set their running times, and add them to soln_group.          */
/*                                                                           */
/*  Implementation note.  Because it may add solutions to soln_group,        */
/*  PInstEndRun may only be called single-threaded.                          */
/*                                                                           */
/*****************************************************************************/

static void PInstEndRun(PINST pinst, KHE_SOLN_GROUP soln_group,
  HA_ARENA_SET final_arena_set, bool set_time)
{
  int i;  KHE_SOLN soln;
  if( soln_group != NULL )
  {
    /* for each solution, reset its arena set, optionally set its running */
    /* time, and add it to soln_group */
    HaArrayForEach(pinst->best_solns, soln, i)
    {
      KheSolnSetArenaSet(soln, final_arena_set);
      if( set_time )
	KheSolnSetRunningTime(soln, pinst->running_time);
      KheSolnGroupAddSoln(soln_group, soln);
    }
  }
  else
  {
    /* delete each solution */
    HaArrayForEach(pinst->best_solns, soln, i)
      KheSolnDelete(soln);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_SOLN PInstEndRunSingle(PINST pinst, HA_ARENA_SET final_arena_set,    */
/*    bool set_time)                                                         */
/*                                                                           */
/*  Similar to PInstEndRun, except there is supposed to be exactly one       */
/*  soln, which is returned.                                                 */
/*                                                                           */
/*****************************************************************************/

static KHE_SOLN PInstEndRunSingle(PINST pinst, HA_ARENA_SET final_arena_set,
  bool set_time)
{
  KHE_SOLN soln;
  HnAssert(HaArrayCount(pinst->best_solns) == 1,
    "PInstEndRunSingle internal error");
  soln = HaArrayFirst(pinst->best_solns);
  if( set_time )
    KheSolnSetRunningTime(soln, pinst->running_time);
  KheSolnSetArenaSet(soln, final_arena_set);
  return soln;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheCostIncreasingCmp(const void *t1, const void *t2)                 */
/*                                                                           */
/*  Comparison function for sorting an array of costs into increasing order. */ 
/*                                                                           */
/*****************************************************************************/

static int KheCostIncreasingCmp(const void *t1, const void *t2)
{
  KHE_COST cost1 = * (KHE_COST *) t1;
  KHE_COST cost2 = * (KHE_COST *) t2;
  return KheCostCmp(cost1, cost2);
}


/*****************************************************************************/
/*                                                                           */
/*  void PInstDebug(PINST pinst, int indent, FILE *fp)                       */
/*                                                                           */
/*  Debug print of pinst.                                                    */
/*                                                                           */
/*****************************************************************************/

static void PInstDebug(PINST pinst, int indent, FILE *fp)
{
  KHE_COST cost;  int i;
  if( HaArrayCount(pinst->all_costs) == 1 )
    fprintf(fp, "%*ssole soln of \"%s\" has cost %.5f\n", indent, "",
      KheInstanceId(pinst->instance),
      KheCostShow(HaArrayFirst(pinst->all_costs)));
  else
  {
    HaArraySort(pinst->all_costs, &KheCostIncreasingCmp);
    HaArraySortUnique(pinst->unique_costs, &KheCostIncreasingCmp);
    fprintf(fp, "%*scosts of all solns of \"%s\" (%d costs, %d unique):",
      indent, "", KheInstanceId(pinst->instance),
      HaArrayCount(pinst->all_costs), HaArrayCount(pinst->unique_costs));
    HaArrayForEach(pinst->all_costs, cost, i)
    {
      if( i % 8 == 0 )
	fprintf(fp, "\n%*s  ", indent, "");
      else
	fprintf(fp, " ");
      fprintf(fp, "%.5f", KheCostShow(cost));
    }
    fprintf(fp, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "PSolver"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  PTHREAD PSolverThreadMake(PSOLVER psolver)                               */
/*                                                                           */
/*  Make a new thread object for psolver.  This thread object does not       */
/*  represent the leader thread, because psolver itself (upcast to PTHREAD)  */
/*  represents the leader thread.                                            */
/*                                                                           */
/*  Implementation note 1.  It would be good to initialize the thread field  */
/*  here to a NULL value, but the pthread module does not seem to supply     */
/*  such a value.  It even says explicitly that you can't apply `==' to the  */
/*  thread type.  Setting the thread running immediately would solve this    */
/*  problem, but that leads to contention on psolver->arena_set.  So we      */
/*  leave the thread field uninitialized.                                    */
/*                                                                           */
/*  Implementation note 2.  Care is needed to avoid contention for memory    */
/*  when running threads.  This is not a problem for PSolverPThreadMake,     */
/*  because it is run single-threaded.  It gives its thread its own arena    */
/*  and its own arena set lying in that arena, and copies psolver->options,  */
/*  so that there can be no contention when the thread runs.                 */
/*                                                                           */
/*****************************************************************************/

static PTHREAD PSolverThreadMake(PSOLVER psolver)
{
  PTHREAD res;  HA_ARENA a;
  a = HaArenaSetArenaBegin(psolver->arena_set);
  HaMake(res, a);
  res->arena = a;
  res->psolver = psolver;
  res->options = (psolver->options == NULL ? NULL :
    KheOptionsCopy(psolver->options, a));
  res->arena_set = (psolver->arena_set == NULL ? NULL : HaArenaSetMake(a));
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolverThreadDelete(PSOLVER psolver, PTHREAD pt)                    */
/*                                                                           */
/*  Delete pt, including merging its arena set's arenas back to psolver.     */
/*                                                                           */
/*****************************************************************************/

static void PSolverThreadDelete(PSOLVER psolver, PTHREAD pt)
{
  if( psolver->arena_set != NULL )
    HaArenaSetMerge(psolver->arena_set, pt->arena_set);
  HaArenaSetArenaEnd(psolver->arena_set, pt->arena);
}


/*****************************************************************************/
/*                                                                           */
/*  PSOLVER PSolverMake(KHE_GENERAL_SOLVER solver, KHE_OPTIONS options,      */
/*    HA_ARENA_SET as, int ps_threads, int ps_make, bool ps_no_diversify,    */
/*    int ps_keep, KHE_PS_TIME ps_time_measure, float ps_time_limit)         */
/*                                                                           */
/*  Make a new psolver object with these attributes, and also with the       */
/*  appropriate number of thread objects, but initially with no instances.   */
/*                                                                           */
/*****************************************************************************/

static PSOLVER PSolverMake(KHE_GENERAL_SOLVER solver, KHE_OPTIONS options,
  HA_ARENA_SET as, int ps_threads, int ps_make, bool ps_no_diversify,
  int ps_keep, KHE_PS_TIME ps_time_measure, float ps_time_limit)
{
  PSOLVER res;  int i;  HA_ARENA a;

  /* fields inherited from PTHREAD */
  a = HaArenaSetArenaBegin(as);
  HaMake(res, a);
  res->arena = a;
  res->psolver = res;
  res->options = options;
  res->arena_set = as;

  /* other fields */
  res->solver = solver;
  res->global_timer = KheStatsTimerMake(a);
  res->ps_make = ps_make;
  res->ps_no_diversify = ps_no_diversify;
  res->ps_keep = ps_keep;
  res->ps_time_measure = ps_time_measure;
  res->ps_time_limit = ps_time_limit;
  res->curr_instance = 0;
  HaArrayInit(res->instances, a);
#if KHE_USE_PTHREAD
  res->ps_threads = ps_threads;
  HaArrayInit(res->other_threads, a);
  for( i = 1;  i < res->ps_threads;  i++ )
    HaArrayAddLast(res->other_threads, PSolverThreadMake(res));
  pthread_mutex_init(&res->mutex, NULL);
#endif
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolverDelete(PSOLVER psolver)                                      */
/*                                                                           */
/*  Delete the psolver and its thread objects, including tidying up the      */
/*  arena sets.                                                              */
/*                                                                           */
/*****************************************************************************/

static void PSolverDelete(PSOLVER psolver)
{
#if KHE_USE_PTHREAD
  PTHREAD pt;  int i;
  HaArrayForEach(psolver->other_threads, pt, i)
    PSolverThreadDelete(psolver, pt);
#endif
  HaArenaSetArenaEnd(psolver->arena_set, psolver->arena);
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolverLock(PSOLVER psolver)                                        */
/*                                                                           */
/*  Lock psolver, or do nothing if we are not running multi-threaded.        */
/*                                                                           */
/*****************************************************************************/

static void PSolverLock(PSOLVER psolver)
{
#if KHE_USE_PTHREAD
  pthread_mutex_lock(&psolver->mutex);
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolverUnLock(PSOLVER psolver)                                      */
/*                                                                           */
/*  Unlock psolver, or do nothing if we are not running multi-threaded.      */
/*                                                                           */
/*****************************************************************************/

static void PSolverUnLock(PSOLVER psolver)
{
#if KHE_USE_PTHREAD
  pthread_mutex_unlock(&psolver->mutex);
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  bool PSolverSolveWanted(PSOLVER psolver, PINST *res, int *diversifier)   */
/*                                                                           */
/*  If there is still work to do, set *res to the pinst needing solving      */
/*  and *diversifier to a suitable diversifier and return true.  Otherwise   */
/*  return false.  Also update psolver.                                      */
/*                                                                           */
/*  Implementation note.  Because psolver is locked during this operation,   */
/*  we prefer to do as little as possible here, i.e. we do not create a      */
/*  new solution.                                                            */
/*                                                                           */
/*****************************************************************************/

static bool PSolverSolveWanted(PSOLVER psolver, PINST *res, int *diversifier)
{
  PINST pinst;
  PSolverLock(psolver);
  *res = NULL;
  *diversifier = -1;
  while( psolver->curr_instance < HaArrayCount(psolver->instances) )
  {
    pinst = HaArray(psolver->instances, psolver->curr_instance);
    if( PInstSolveWanted(pinst, diversifier) )
    {
      *res = pinst;
      break;
    }
    psolver->curr_instance++;
  }
  PSolverUnLock(psolver);
  return *res != NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolverSolveEnded(PSOLVER psolver, PINST pinst, KHE_SOLN soln,      */
/*    KHE_SOLN *unwanted_soln)                                               */
/*                                                                           */
/*  Add soln to pinst of psolver.  If *unwanted_soln is non-NULL on return,  */
/*  it is not wanted and should be deleted.                                  */
/*                                                                           */
/*  Strictly speaking we only need to lock pinst while this is going on,     */
/*  but given that two solves that finish around the same time are likely    */
/*  to be for the same instance, it makes little difference if we lock the   */
/*  whole solver.                                                            */
/*                                                                           */
/*****************************************************************************/

static void PSolverSolveEnded(PSOLVER psolver, PINST pinst, KHE_SOLN soln,
  KHE_SOLN *unwanted_soln)
{
  PSolverLock(psolver);
  PInstSolveEnded(pinst, soln, unwanted_soln);
  PSolverUnLock(psolver);
}


/*****************************************************************************/
/*                                                                           */
/*  void *PSolverRunOneThread(void *arg)                                     */
/*                                                                           */
/*  Thread execution function.                                               */
/*                                                                           */
/*  Implementation note.  All memory allocated by this function comes from   */
/*  the thread's own arena and arena set.                                    */
/*                                                                           */
/*****************************************************************************/

static void *PSolverRunOneThread(void *arg)
{
  PSOLVER psolver;  PINST pinst;  KHE_SOLN soln, unwanted_soln;
  int diversifier;  KHE_OPTIONS options;  PTHREAD pt;  char description[20];
  pt = (PTHREAD) arg;
  psolver = pt->psolver;
  while( PSolverSolveWanted(psolver, &pinst, &diversifier) )
  {
    soln = KheSolnMake(pinst->instance, pt->arena_set);
    if( !psolver->ps_no_diversify )
      KheSolnSetDiversifier(soln, diversifier);
    options = (pt->options == NULL ? NULL :
      KheOptionsCopy(pt->options, pt->arena));
    soln = psolver->solver(soln, options);
    sprintf(description, "diversifier %d", KheSolnDiversifier(soln));
    KheSolnSetDescription(soln, description);  /* will copy description */
    PSolverSolveEnded(psolver, pinst, soln, &unwanted_soln);
    if( unwanted_soln != NULL )
      KheSolnDelete(unwanted_soln);
  }
  return NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolverRun(PSOLVER psolver)                                         */
/*                                                                           */
/*  Run psolver (called by the leader thread).                               */
/*                                                                           */
/*****************************************************************************/

static void PSolverRun(PSOLVER psolver)
{
#if KHE_USE_PTHREAD
  int i;  PTHREAD pt;

  /* set the other threads running from the first instance */
  psolver->curr_instance = 0;
  HaArrayForEach(psolver->other_threads, pt, i)
    pthread_create(&pt->thread, NULL, &PSolverRunOneThread, (void *) pt);

  /* run this thread */
  psolver->thread = pthread_self();  /* actually unused */
  PSolverRunOneThread((void *) (PTHREAD) psolver);

  /* wait for the other threads to terminate */
  HaArrayForEach(psolver->other_threads, pt, i)
    pthread_join(pt->thread, NULL);
#else
  PSolverRunOneThread((void *) (PTHREAD) psolver);
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolverLoadInstance(PSOLVER psolver, KHE_INSTANCE ins)              */
/*                                                                           */
/*  Load ins into psolver.                                                   */
/*                                                                           */
/*****************************************************************************/

static PINST PSolverLoadInstance(PSOLVER psolver, KHE_INSTANCE ins)
{
  PINST res;
  res = PInstMake(ins, psolver->ps_make, psolver->ps_time_limit,
      psolver->ps_keep, psolver->arena);
  HaArrayAddLast(psolver->instances, res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "public functions"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheGetPSOptions(KHE_OPTIONS options, bool keep_is_one,              */
/*    int *ps_threads, int *ps_make, bool *ps_no_diversify, int *ps_keep,    */
/*    KHE_PS_TIME *ps_time_measure, float *ps_time_limit)                    */
/*                                                                           */
/*  Get the parallel solve options from options.  If keep_is_one, set        */
/*  ps_keep to 1 regardless of what options says.                            */
/*                                                                           */
/*****************************************************************************/

static void KheGetPSOptions(KHE_OPTIONS options, bool keep_is_one,
  int *ps_threads, int *ps_make, bool *ps_no_diversify, int *ps_keep,
  KHE_PS_TIME *ps_time_measure, float *ps_time_limit)
{
  char *str;
  *ps_threads = KheOptionsGetInt(options, "ps_threads", 1);
  HnAssert(*ps_threads >= 1, "KheParallelSolve: ps_threads < 1");
  *ps_make = KheOptionsGetInt(options, "ps_make", 1);
  HnAssert(*ps_make >= 1, "KheParallelSolve: ps_make < 1");
  *ps_no_diversify = KheOptionsGetBool(options, "ps_no_diversify", false);
  *ps_keep = (keep_is_one ? 1 : KheOptionsGetInt(options, "ps_keep", 1));
  HnAssert(*ps_keep >= 0, "KheParallelSolve: ps_keep < 0");
  str = KheOptionsGet(options, "ps_time_measure", "auto");
  if( strcmp(str, "auto") == 0 )
    *ps_time_measure = (*ps_keep < *ps_make ?
      KHE_PS_TIME_SHARED : KHE_PS_TIME_OMIT);
  else if( strcmp(str, "omit") == 0 )
    *ps_time_measure = KHE_PS_TIME_OMIT;
  else if( strcmp(str, "shared") == 0 )
    *ps_time_measure = KHE_PS_TIME_SHARED;
  else
  {
    HnAbort("KheParallelSolve: ps_time_measure option has unknown value \"%s\"",
      str);
    *ps_time_measure = KHE_PS_TIME_OMIT;  /* keep compiler happy */
  }
  *ps_time_limit =
    KheConvertTimeLimit(KheOptionsGet(options, "ps_time_limit", "-"));
}


/*****************************************************************************/
/*                                                                           */
/*  char *ElapsedTime(float secs)                                            */
/*                                                                           */
/*  Return a string display of the elapsed time, in static memory.           */
/*                                                                           */
/*****************************************************************************/

static char *ElapsedTime(float secs)
{
  static char res[50];
  if( secs > 300.0 )
    sprintf(res, "%.1f mins", secs / 60.0);
  else
    sprintf(res, "%.1f secs", secs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveParallelSolve(KHE_ARCHIVE archive,                        */
/*    KHE_GENERAL_SOLVER solver, KHE_OPTIONS options, HA_ARENA_SET as,       */
/*    KHE_SOLN_GROUP soln_group)                                             */
/*                                                                           */
/*  Solve the instances of archive in parallel, and, if soln_group is not    */
/*  NULL, add the best solutions to soln_group.                              */
/*                                                                           */
/*****************************************************************************/

void KheArchiveParallelSolve(KHE_ARCHIVE archive, KHE_GENERAL_SOLVER solver,
  KHE_OPTIONS options, HA_ARENA_SET as, KHE_SOLN_GROUP soln_group)
{
  PSOLVER psolver;  int i;  PINST pinst;  KHE_SOLN soln;
  KHE_INSTANCE ins;  KHE_PS_TIME ps_time_measure;
  float ps_time_limit; int ps_threads, ps_make, ps_keep;
  bool ps_no_diversify;

  /* get options */
  KheGetPSOptions(options, false, &ps_threads, &ps_make, &ps_no_diversify,
    &ps_keep, &ps_time_measure, &ps_time_limit);
  if( DEBUG2 )
    fprintf(stderr, "[ KheArchiveParallelSolve(%s, threads %d, make %d, "
      "keep %d, time %s, limit %.1f)\n",
      KheArchiveId(archive) == NULL ? "unnamed archive" : KheArchiveId(archive),
      ps_threads, ps_make, ps_keep,
      ps_time_measure == KHE_PS_TIME_OMIT ? "omit" : "shared", ps_time_limit);

  /* create a psolver with ps_threads threads */
  psolver = PSolverMake(solver, options, as, ps_threads, ps_make,
    ps_no_diversify, ps_keep, ps_time_measure, ps_time_limit);

  /* solve the archive, either all together or instance by instance */
  switch( ps_time_measure )
  {
    case KHE_PS_TIME_OMIT:

      /* load and solve the whole archive */
      for( i = 0;  i < KheArchiveInstanceCount(archive);  i++ )
      {
	ins = KheArchiveInstance(archive, i);
	PSolverLoadInstance(psolver, ins);
      }
      PSolverRun(psolver);

      /* add the surviving solutions to soln_group, if any */
      HaArrayForEach(psolver->instances, pinst, i)
	PInstEndRun(pinst, soln_group, psolver->arena_set, false);
      HaArrayClear(psolver->instances);
      break;

    case KHE_PS_TIME_SHARED:

      for( i = 0;  i < KheArchiveInstanceCount(archive);  i++ )
      {
	/* load and solve the i'th instance of archive */
	ins = KheArchiveInstance(archive, i);
	pinst = PSolverLoadInstance(psolver, ins);
	PSolverRun(psolver);

	/* add the surviving solutions to soln_group, if any; set run times */
	PInstEndRun(pinst, soln_group, psolver->arena_set, true);
	HaArrayClear(psolver->instances);
      }
      break;

    default:

      HnAbort("KheArchiveParallelSolve: internal error 1");
      break;
  }

  /* debug prints */
  if( DEBUG1 )
  {
    if( soln_group != NULL && KheSolnGroupSolnCount(soln_group) != 0 )
    {
      fprintf(stderr, "  KheArchiveParallelSolve best solution%s:\n",
	KheSolnGroupSolnCount(soln_group) != 1 ? "s" : "");
      for( i = 0;  i < KheSolnGroupSolnCount(soln_group);  i++ )
      {
	soln = KheSolnGroupSoln(soln_group, i);
	KheSolnDebug(soln, 2, 2, stderr);
      }
    }
  }
  if( DEBUG2 )
    fprintf(stderr, "] KheArchiveParallelSolve returning (%s elapsed)\n",
      ElapsedTime(KheStatsTimerNow(psolver->global_timer)));

  /* delete the solver, including tidying up arena sets */
  PSolverDelete(psolver);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_SOLN KheInstanceParallelSolve(KHE_INSTANCE ins,                      */
/*    KHE_GENERAL_SOLVER solver, KHE_OPTIONS options, HA_ARENA_SET as)       */
/*                                                                           */
/*  Like KheArchiveParallelSolve except that it solves only one instance     */
/*  and saves (and returns) any one best solution.                           */
/*                                                                           */
/*****************************************************************************/

KHE_SOLN KheInstanceParallelSolve(KHE_INSTANCE ins,
  KHE_GENERAL_SOLVER solver, KHE_OPTIONS options, HA_ARENA_SET as)
{
  PSOLVER psolver;  PINST pinst;  KHE_SOLN soln;
  int ps_threads, ps_make, ps_keep;  bool ps_no_diversify;
  KHE_PS_TIME ps_time_measure;  float ps_time_limit;

  /* get options */
  KheGetPSOptions(options, true, &ps_threads, &ps_make, &ps_no_diversify,
    &ps_keep, &ps_time_measure, &ps_time_limit);
  if( DEBUG2 )
    fprintf(stderr, "[ KheInstanceParallelSolve(threads %d, make %d)\n",
      ps_threads, ps_make);

  /* create a psolver object and load and solve ins */
  psolver = PSolverMake(solver, options, as, ps_threads, ps_make,
    ps_no_diversify, 1, ps_time_measure, ps_time_limit);
  pinst = PSolverLoadInstance(psolver, ins);
  PSolverRun(psolver);

  /* extract the sole surviving solution */
  soln = PInstEndRunSingle(pinst, psolver->arena_set,
    ps_time_measure == KHE_PS_TIME_SHARED);
  if( DEBUG2 )
  {
    fprintf(stderr, "  final solution:\n");
    KheSolnDebug(soln, 2, 2, stderr);
    fprintf(stderr, "] KheInstanceParallelSolve returning\n");
  }
  return soln;
}
