
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sm_parallel_solve.c                                    */
/*  DESCRIPTION:  KheParallelSolve().  If you can't compile this file,       */
/*                see the makefile for a workaround.                         */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include "howard_a.h"
#include "howard_n.h"
#include <limits.h>
#include <math.h>
#if KHE_USE_PTHREAD
#include <pthread.h>
#endif

#define DEBUG1 0
#define DEBUG2 1
#define DEBUG3 0


/*****************************************************************************/
/*                                                                           */
/*  PINST - an instance to be solved in parallel                             */
/*                                                                           */
/*  Keeping track of where we are up to with an instance is quite tricky.    */
/*  We use three variables:                                                  */
/*                                                                           */
/*    wanted_solves                                                          */
/*      The number of solves of pinst->instance that we want to do.          */
/*      Initially this will be psolve->ps_make, but if time runs out         */
/*      it is reduced to initiated_solves.                                   */
/*                                                                           */
/*    initiated_solves                                                       */
/*      The number of solves of pinst->instance that have been initiated.    */
/*                                                                           */
/*    completed_solves                                                       */
/*      The number of solves of pinst->instance that have been completed.    */
/*      This is always equal to HaArrayCount(pinst->all_costs).              */
/*                                                                           */
/*  Invariant: wanted_solves >= initiated_solves >= completed_solves         */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_SOLN) ARRAY_KHE_SOLN;
typedef HA_ARRAY(KHE_COST) ARRAY_KHE_COST;

typedef struct pinst_rec {
  KHE_INSTANCE			instance;
  KHE_STATS_TIMER		timer;
  float				running_time;
  ARRAY_KHE_SOLN		best_solns;
  int				wanted_solves;
  int				initiated_solves;
  int				completed_solves;
  ARRAY_KHE_COST		all_costs;
  ARRAY_KHE_COST		unique_costs;
} *PINST;

typedef HA_ARRAY(PINST) ARRAY_PINST;
#if KHE_USE_PTHREAD
typedef HA_ARRAY(pthread_t) ARRAY_PTHREAD;
#endif


/*****************************************************************************/
/*                                                                           */
/*  PSOLVE - a parallel solve                                                */
/*                                                                           */
/*****************************************************************************/

typedef enum {
  KHE_PS_TIME_OMIT,
  KHE_PS_TIME_SHARED
} KHE_PS_TIME;

typedef HA_ARRAY(HA_ARENA_SET) ARRAY_HA_ARENA_SET;

typedef struct psolve_rec {
#if KHE_USE_PTHREAD
  pthread_mutex_t	mutex;			/* locks entire record       */
  pthread_t		leader_thread;		/* leader thread             */
  ARRAY_PTHREAD		other_threads;		/* other threads             */
#endif
  HA_ARENA		arena;			/* holds options etc.        */
  ARRAY_HA_ARENA_SET	arena_sets;		/* arena sets                */
  KHE_ARCHIVE		archive;		/* archive parameter         */
  int			ps_threads;		/* ps_threads option         */
  int			ps_make;		/* ps_make option            */
  bool			ps_no_diversify;	/* ps_no_diversify option    */
  int			ps_keep;		/* ps_keep option            */
  KHE_PS_TIME		ps_time_measure;	/* ps_time_measure option    */
  float			ps_time_limit;		/* time limit, or -1 if none */
  KHE_GENERAL_SOLVER	solver;			/* solver parameter          */
  KHE_OPTIONS		options;		/* solver options            */
  int			curr_instance;		/* currently on this         */
  ARRAY_PINST		instances;		/* instances being solved    */
} *PSOLVE;


/*****************************************************************************/
/*                                                                           */
/*  PINST PInstMake(KHE_INSTANCE ins, int wanted_solves, HA_ARENA a)         */
/*                                                                           */
/*  Make a new pinst object for ins.                                         */
/*                                                                           */
/*****************************************************************************/

static PINST PInstMake(KHE_INSTANCE ins, int wanted_solves, HA_ARENA a)
{
  PINST res;
  HaMake(res, a);
  res->instance = ins;
  res->timer = KheStatsTimerMake(a);
  res->running_time = 0.0;
  HaArrayInit(res->best_solns, a);
  res->wanted_solves = wanted_solves;
  res->initiated_solves = 0;
  res->completed_solves = 0;
  HaArrayInit(res->all_costs, a);
  HaArrayInit(res->unique_costs, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void PInstFree(PINST pinst)                                              */
/*                                                                           */
/*  Free pinst.                                                              */
/*                                                                           */
/*****************************************************************************/

/* ***
static void PInstFree(PINST pinst)
{
  KheStatsTimerDelete(pinst->timer);
  MArrayFree(pinst->best_solns);
  MFree(pinst);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  PSOLVE PSolveMake(KHE_ARCHIVE archive, KHE_INSTANCE ins,                 */
/*    int ps_threads, int ps_make, bool ps_no_diversify, int ps_keep,        */
/*    KHE_PS_TIME ps_time_measure, float ps_time_limit,                      */
/*    KHE_GENERAL_SOLVER solver, KHE_OPTIONS options, HA_ARENA a)            */
/*                                                                           */
/*  Make a new psolve object with these attributes.                          */
/*                                                                           */
/*  The psolve object does not consult options itself, it merely passes      */
/*  it on to solver.                                                         */
/*                                                                           */
/*****************************************************************************/

static PSOLVE PSolveMake(KHE_ARCHIVE archive, KHE_INSTANCE ins,
  int ps_threads, int ps_make, bool ps_no_diversify, int ps_keep,
  KHE_PS_TIME ps_time_measure, float ps_time_limit,
  KHE_GENERAL_SOLVER solver, KHE_OPTIONS options, HA_ARENA a)
{
  PSOLVE res;  int i;
  HaMake(res, a);
#if KHE_USE_PTHREAD
  pthread_mutex_init(&res->mutex, NULL);
  res->leader_thread = pthread_self();
  HaArrayInit(res->other_threads, a);
#endif
  res->arena = a;
  HaArrayInit(res->arena_sets, a);
  res->archive = archive;
  res->ps_threads = ps_threads;
  res->ps_make = ps_make;
  res->ps_no_diversify = ps_no_diversify;
  res->ps_keep = ps_keep;
  res->ps_time_measure = ps_time_measure;
  res->ps_time_limit = ps_time_limit;
  res->solver = solver;
  res->options = options;
  res->curr_instance = 0;
  HaArrayInit(res->instances, a);
  if( archive != NULL )
    for( i = 0;  i < KheArchiveInstanceCount(archive);  i++ )
    {
      ins = KheArchiveInstance(archive, i);
      /* if( KheBenchmarkTryInstance(ins) ) */
	HaArrayAddLast(res->instances, PInstMake(ins, ps_make, a));
    }
  else
  {
    /* if( KheBenchmarkTryInstance(ins) ) */
      HaArrayAddLast(res->instances, PInstMake(ins, ps_make, a));
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolveFree(PSOLVE psolve)                                           */
/*                                                                           */
/*  Free psolve.                                                             */
/*                                                                           */
/*****************************************************************************/

/* ***
static void PSolveFree(PSOLVE psolve)
{
#if KHE_USE_PTHREAD
  MArrayFree(psolve->other_threads);
#endif
  while( HaArrayCount(psolve->instances) > 0 )
    PInstFree(HaArrayLastAndDelete(psolve->instances));
  MArrayFree(psolve->instances);
  MFree(psolve);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KhePSolveTimeAvailable(PSOLVE psolve, PINST pinst)                  */
/*                                                                           */
/*  Return true if there is still time available to solve pinst.             */
/*                                                                           */
/*****************************************************************************/

static bool KhePSolveTimeAvailable(PSOLVE psolve, PINST pinst)
{
  return psolve->ps_time_limit == -1 ||
    KheStatsTimerNow(pinst->timer) < psolve->ps_time_limit;
}


/*****************************************************************************/
/*                                                                           */
/*  bool PSolveBeginSolve(PSOLVE psolve, PINST *res, int *diversifier)       */
/*                                                                           */
/*  If there is still work to do, set *res to the pinst needing solving      */
/*  and *diversifier to a suitable diversifier and return true.  Otherwise   */
/*  return false.  Also update psolve.                                       */
/*                                                                           */
/*****************************************************************************/
static void PInstDebug(PINST pinst, int indent, FILE *fp);

static bool PSolveBeginSolve(PSOLVE psolve, PINST *res, int *diversifier,
  KHE_OPTIONS *options, HA_ARENA_SET *as)
{
  PINST pinst;
#if KHE_USE_PTHREAD
  pthread_mutex_lock(&psolve->mutex);
#endif
  while( psolve->curr_instance < HaArrayCount(psolve->instances) )
  {
    pinst = HaArray(psolve->instances, psolve->curr_instance);
    if( pinst->initiated_solves < pinst->wanted_solves )
    {
      if( pinst->initiated_solves == 0 )
      {
	KheStatsTimerReset(pinst->timer);
	if( DEBUG2 )
	  fprintf(stderr, "  starting first of up to %d solves of %s:\n",
            psolve->ps_make, KheInstanceId(pinst->instance));
      }
      if( KhePSolveTimeAvailable(psolve, pinst) )
      {
	*res = pinst;
	*diversifier = (psolve->ps_no_diversify ? 0 : pinst->initiated_solves);
	*options = psolve->options == NULL ? NULL :
	  KheOptionsCopy(psolve->options, psolve->arena);
	if( HaArrayCount(psolve->arena_sets) > 0 )
	  *as = HaArrayLastAndDelete(psolve->arena_sets);
	else
	  *as = HaArenaSetMake(psolve->arena);
	pinst->initiated_solves++;
	if( DEBUG3 )
	  fprintf(stderr, "  PSolveBeginSolve %s: initiating solve %d\n",
	    KheInstanceId(pinst->instance), pinst->initiated_solves);
#if KHE_USE_PTHREAD
	pthread_mutex_unlock(&psolve->mutex);
#endif
	return true;
      }
      else
      {
	pinst->wanted_solves = pinst->initiated_solves;
        if( DEBUG3 )
	  fprintf(stderr, "  PSolveBeginSolve %s: only %d of %d solves "
	    "initiated (out of time)\n", KheInstanceId(pinst->instance),
	    pinst->initiated_solves, psolve->ps_make);
      }
    }
    else
      if( DEBUG3 )
	fprintf(stderr, "  PSolveBeginSolve %s: all %d solves initiated\n",
	  KheInstanceId(pinst->instance), psolve->ps_make);
    /* *** this is really "initiated last" not "ended last"
    if( DEBUG2 )
    {
      fprintf(stderr, "  ended last of %d solves of %s:\n",
	pinst->initiated_solves, KheInstanceId(pinst->instance));
      PInstDebug(pinst, 4, stderr);
    }
    *** */
    psolve->curr_instance++;
  }
#if KHE_USE_PTHREAD
  pthread_mutex_unlock(&psolve->mutex);
#endif
  if( DEBUG3 )
    fprintf(stderr, "  PSolveBeginSolve: all solves initiated\n");
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolveEndSolve(PSOLVE psolve, PINST pinst, KHE_SOLN soln)           */
/*                                                                           */
/*  Add soln to pinst of psolve, and possibly remove and delete the worst.   */
/*                                                                           */
/*****************************************************************************/

static void PSolveEndSolve(PSOLVE psolve, PINST pinst, KHE_SOLN soln,
  HA_ARENA_SET as)
{
  int i;  KHE_SOLN best_soln;
#if KHE_USE_PTHREAD
  pthread_mutex_lock(&psolve->mutex);
#endif
  pinst->completed_solves++;
  HaArrayAddLast(psolve->arena_sets, as);
  HaArrayAddLast(pinst->all_costs, KheSolnCost(soln));
  HaArrayAddLast(pinst->unique_costs, KheSolnCost(soln));
  for( i = 0;  i < HaArrayCount(pinst->best_solns);  i++ )
  {
    best_soln = HaArray(pinst->best_solns, i);
    if( KheSolnCost(soln) < KheSolnCost(best_soln) )
    {
      HaArrayAdd(pinst->best_solns, i, soln);
      break;
    }
  }
  if( i >= HaArrayCount(pinst->best_solns) )
    HaArrayAddLast(pinst->best_solns, soln);
  if( HaArrayCount(pinst->best_solns) > psolve->ps_keep )
    KheSolnDelete(HaArrayLastAndDelete(pinst->best_solns));
  pinst->running_time = KheStatsTimerNow(pinst->timer);
  if( pinst->completed_solves == pinst->wanted_solves )
  {
    fprintf(stderr, "  ended last of %d solves of %s:\n",
      pinst->completed_solves, KheInstanceId(pinst->instance));
    PInstDebug(pinst, 4, stderr);
  }
#if KHE_USE_PTHREAD
    pthread_mutex_unlock(&psolve->mutex);
#endif
}


/*****************************************************************************/
/*                                                                           */
/*  void *PSolveRun(void *arg)                                               */
/*                                                                           */
/*  Thread execution function.                                               */
/*                                                                           */
/*****************************************************************************/

static void *PSolveRun(void *arg)
{
  PSOLVE psolve;  PINST pinst;  KHE_SOLN soln;  int diversifier;
  HA_ARENA_SET as;  /* HA_ARENA a;  int i; */  KHE_OPTIONS options;
  /*  as = HaArenaSetMake(a); */
  psolve = (PSOLVE) arg;
  while( PSolveBeginSolve(psolve, &pinst, &diversifier, &options, &as) )
  {
    soln = KheSolnMake(pinst->instance, as);
    KheSolnSetDiversifier(soln, diversifier);
    soln = psolve->solver(soln, options);
    /* ***
    soln = psolve->solver(soln, psolve->options == NULL ? NULL :
      KheOptionsCopy(psolve->options));
    *** */
    PSolveEndSolve(psolve, pinst, soln, as);
  }
  /* ***
  for( i = 0;  i < HaArenaSetArenaCount(as);  i++ )
    HaArenaDelete(HaArenaSetArena(as, i));
  HaArenaDelete(a);
  *** */
  return NULL;
}


/*****************************************************************************/
/*                                                                           */
/*  void PSolveRunAll(PSOLVE psolve)                                         */
/*                                                                           */
/*  Run psolve completely (called by the leader thread).                     */
/*                                                                           */
/*****************************************************************************/

static void PSolveRunAll(PSOLVE psolve)
{
  int i;
#if KHE_USE_PTHREAD
  pthread_t thread;
#endif

  /* start the other threads */
#if KHE_USE_PTHREAD
  pthread_mutex_lock(&psolve->mutex);
  for( i = 1;  i < psolve->ps_threads;  i++ )
  {
    pthread_create(&thread, NULL, &PSolveRun, psolve);
    HaArrayAddLast(psolve->other_threads, thread);
  }
  pthread_mutex_unlock(&psolve->mutex);
#endif

  /* run this thread */
  PSolveRun((void *) psolve);

  /* wait for the other threads to terminate */
#if KHE_USE_PTHREAD
  HaArrayForEach(psolve->other_threads, thread, i)
    pthread_join(thread, NULL);
#endif

  /* free memory, if requested */
  /* ***
  if( psolve->ps_free_memory )
    HaArrayForEach(psolve->arena_sets, as, i)
      for( j = 0;  j < HaArenaSetArenaCount(as);  j++ )
	HaArenaDelete(HaArenaSetArena(as, j));
  *** */
}


/*****************************************************************************/
/*                                                                           */
/*  int KheCostIncreasingCmp(const void *t1, const void *t2)                 */
/*                                                                           */
/*  Comparison function for sorting an array of costs into increasing order. */ 
/*                                                                           */
/*****************************************************************************/

static int KheCostIncreasingCmp(const void *t1, const void *t2)
{
  KHE_COST cost1 = * (KHE_COST *) t1;
  KHE_COST cost2 = * (KHE_COST *) t2;
  return KheCostCmp(cost1, cost2);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGetPSOptions(KHE_OPTIONS options, bool keep_is_one,              */
/*    int *ps_threads, int *ps_make, bool *ps_no_diversify, int *ps_keep,    */
/*    KHE_PS_TIME *ps_time_measure, float *ps_time_limit)                    */
/*                                                                           */
/*  Get the parallel solve options from options.  If keep_is_one, set        */
/*  ps_keep to 1 regardless of what options says.                            */
/*                                                                           */
/*****************************************************************************/

static void KheGetPSOptions(KHE_OPTIONS options, bool keep_is_one,
  int *ps_threads, int *ps_make, bool *ps_no_diversify, int *ps_keep,
  KHE_PS_TIME *ps_time_measure, float *ps_time_limit)
{
  char *str;
  *ps_threads = KheOptionsGetInt(options, "ps_threads", 1);
  HnAssert(*ps_threads >= 1, "KheParallelSolve: ps_threads < 1");
  *ps_make = KheOptionsGetInt(options, "ps_make", 1);
  HnAssert(*ps_make >= 1, "KheParallelSolve: ps_make < 1");
  *ps_no_diversify = KheOptionsGetBool(options, "ps_no_diversify", false);
  *ps_keep = (keep_is_one ? 1 : KheOptionsGetInt(options, "ps_keep", 1));
  HnAssert(*ps_keep >= 0, "KheParallelSolve: ps_keep < 0");
  str = KheOptionsGet(options, "ps_time_measure", "auto");
  if( strcmp(str, "auto") == 0 )
    *ps_time_measure = (*ps_keep < *ps_make ?
      KHE_PS_TIME_SHARED : KHE_PS_TIME_OMIT);
  else if( strcmp(str, "omit") == 0 )
    *ps_time_measure = KHE_PS_TIME_OMIT;
  else if( strcmp(str, "shared") == 0 )
    *ps_time_measure = KHE_PS_TIME_SHARED;
  else
  {
    HnAbort("KheParallelSolve: ps_time_measure option has unknown value \"%s\"",
      str);
    *ps_time_measure = KHE_PS_TIME_OMIT;  /* keep compiler happy */
  }
  *ps_time_limit =
    KheConvertTimeLimit(KheOptionsGet(options, "ps_time_limit", "-"));
}


/*****************************************************************************/
/*                                                                           */
/*  void PInstDebug(PINST pinst, int indent, FILE *fp)                       */
/*                                                                           */
/*  Debug print of pinst.                                                    */
/*                                                                           */
/*****************************************************************************/

static void PInstDebug(PINST pinst, int indent, FILE *fp)
{
  KHE_COST cost;  int i;
  if( HaArrayCount(pinst->all_costs) == 1 )
    fprintf(fp, "%*ssole soln of \"%s\" has cost %.5f\n", indent, "",
      KheInstanceId(pinst->instance),
      KheCostShow(HaArrayFirst(pinst->all_costs)));
  else
  {
    HaArraySort(pinst->all_costs, &KheCostIncreasingCmp);
    HaArraySortUnique(pinst->unique_costs, &KheCostIncreasingCmp);
    fprintf(fp, "%*scosts of all solns of \"%s\" (%d costs, %d unique):",
      indent, "", KheInstanceId(pinst->instance),
      HaArrayCount(pinst->all_costs), HaArrayCount(pinst->unique_costs));
    HaArrayForEach(pinst->all_costs, cost, i)
    {
      if( i % 8 == 0 )
	fprintf(fp, "\n%*s  ", indent, "");
      else
	fprintf(fp, " ");
      fprintf(fp, "%.5f", KheCostShow(cost));
    }
    fprintf(fp, "\n");
    /* ***
    fprintf(fp, "%*sunique costs of solns of \"%s\" (%d):", indent, "",
      KheInstanceId(pinst->instance), HaArrayCount(pinst->all_costs));
    HaArrayForEach(pinst->all_costs, cost, i)
    {
      if( i % 8 == 0 )
	fprintf(fp, "\n%*s  ", indent, "");
      else
	fprintf(fp, " ");
      fprintf(fp, "%.5f", KheCostShow(cost));
    }
    fprintf(fp, "\n");
    *** */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *ElapsedTime(float secs)                                            */
/*                                                                           */
/*  Return a string display of the elapsed time, in static memory.           */
/*                                                                           */
/*****************************************************************************/

static char *ElapsedTime(float secs)
{
  static char res[50];
  if( secs > 300.0 )
    sprintf(res, "%.1f mins", secs / 60.0);
  else
    sprintf(res, "%.1f secs", secs);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheArchiveParallelSolve(KHE_ARCHIVE archive,                        */
/*    KHE_GENERAL_SOLVER solver, HA_ARENA_SET as, KHE_OPTIONS options,       */
/*    KHE_SOLN_GROUP soln_group)                                             */
/*                                                                           */
/*  Solve the instances of archive in parallel, and, if soln_group is not    */
/*  NULL, add the best solutions to soln_group.                              */
/*                                                                           */
/*****************************************************************************/

void KheArchiveParallelSolve(KHE_ARCHIVE archive, KHE_GENERAL_SOLVER solver,
  HA_ARENA_SET as, KHE_OPTIONS options, KHE_SOLN_GROUP soln_group)
{
  PSOLVE psolve;  int i, j;  PINST pinst;  KHE_SOLN soln;
  KHE_STATS_TIMER st;  KHE_INSTANCE ins;  KHE_PS_TIME ps_time_measure;
  float ps_time_limit; int ps_threads, ps_make, ps_keep;
  bool ps_no_diversify;  HA_ARENA a;

  /* get options */
  a = HaArenaMake();
  KheGetPSOptions(options, false, &ps_threads, &ps_make, &ps_no_diversify,
    &ps_keep, &ps_time_measure, &ps_time_limit);
  if( DEBUG2 )
  {
    fprintf(stderr, "[ KheArchiveParallelSolve(%s, threads %d, make %d, "
      "keep %d, time %s, limit %.1f)\n", KheArchiveId(archive),
      ps_threads, ps_make, ps_keep,
      ps_time_measure == KHE_PS_TIME_OMIT ? "omit" : "shared", ps_time_limit);
    st = KheStatsTimerMake(a);
  }

  switch( ps_time_measure )
  {
    case KHE_PS_TIME_OMIT:

      /* create a psolve object for the entire archive */
      psolve = PSolveMake(archive, NULL, ps_threads, ps_make, ps_no_diversify,
	ps_keep, ps_time_measure, ps_time_limit, solver, options, a);

      /* do the actual parallel solve */
      PSolveRunAll(psolve);

      /* add the surviving solutions to soln_group, if any */
      if( soln_group != NULL )
	HaArrayForEach(psolve->instances, pinst, i)
	  HaArrayForEach(pinst->best_solns, soln, j)
	    KheSolnGroupAddSoln(soln_group, soln);
      /* ***
      if( DEBUG2 )
	HaArrayForEach(psolve->instances, pinst, i)
	  PInstDebug(pinst, 2, stderr);
      *** */
      /* PSolveFree(psolve); */
      break;

    case KHE_PS_TIME_SHARED:

      for( i = 0;  i < KheArchiveInstanceCount(archive);  i++ )
      {
	ins = KheArchiveInstance(archive, i);

	/* create a psolve object for ins only */
	/* KheStatsTimerReset(st); */
	psolve = PSolveMake(NULL, ins, ps_threads, ps_make, ps_no_diversify,
	  ps_keep, ps_time_measure, ps_time_limit, solver, options, a);

	/* do the actual parallel solve */
	PSolveRunAll(psolve);

	/* add the surviving solutions to soln_group, if any; set run times */
	HnAssert(HaArrayCount(psolve->instances) == 1,
	  "KheArchiveParallelSolve internal error");
	pinst = HaArrayFirst(psolve->instances);
	/* secs = pinst->running_time; */
	/* secs = KheStatsTimerNow(st); */
	if( soln_group != NULL )
	  /* HaArrayForEach(psolve->instances, pinst, j) */
	    HaArrayForEach(pinst->best_solns, soln, j)
	    {
	      KheSolnSetRunningTime(soln, pinst->running_time);
	      KheSolnGroupAddSoln(soln_group, soln);
	    }
	/* ***
	if( DEBUG2 )
	  ** HaArrayForEach(psolve->instances, pinst, j) **
	    PInstDebug(pinst, 2, stderr);
	*** */

	/* delete the psolve object */
	/* PSolveFree(psolve); */
      }
      break;

    default:

      HnAbort("KheArchiveParallelSolve: internal error 1");
      break;
  }
  if( DEBUG1 )
  {
    if( soln_group != NULL && KheSolnGroupSolnCount(soln_group) != 0 )
    {
      fprintf(stderr, "  KheArchiveParallelSolve best solution%s:\n",
	KheSolnGroupSolnCount(soln_group) != 1 ? "s" : "");
      for( i = 0;  i < KheSolnGroupSolnCount(soln_group);  i++ )
      {
	soln = KheSolnGroupSoln(soln_group, i);
	KheSolnDebug(soln, 2, 2, stderr);
      }
    }
  }
  if( DEBUG2 )
  {
    fprintf(stderr, "] KheArchiveParallelSolve returning (%s elapsed)\n",
      ElapsedTime(KheStatsTimerNow(st)));
    /* KheStatsTimerDelete(st); */
  }
  HaArenaDelete(a);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_SOLN KheInstanceParallelSolve(KHE_INSTANCE ins,                      */
/*    KHE_GENERAL_SOLVER solver, HA_ARENA_SET as, KHE_OPTIONS options)       */
/*                                                                           */
/*  Like KheArchiveParallelSolve except that it solves only one instance     */
/*  and saves (and returns) any one best solution.                           */
/*                                                                           */
/*****************************************************************************/

KHE_SOLN KheInstanceParallelSolve(KHE_INSTANCE ins,
  KHE_GENERAL_SOLVER solver, HA_ARENA_SET as, KHE_OPTIONS options)
{
  PSOLVE psolve;  PINST pinst;  KHE_SOLN soln; KHE_STATS_TIMER st;
  int ps_threads, ps_make, ps_keep;  bool ps_no_diversify;
  KHE_PS_TIME ps_time_measure;  float ps_time_limit;  HA_ARENA a;

  /* get options */
  KheGetPSOptions(options, true, &ps_threads, &ps_make, &ps_no_diversify,
    &ps_keep, &ps_time_measure, &ps_time_limit);
  if( DEBUG2 )
    fprintf(stderr, "[ KheInstanceParallelSolve(threads %d, make %d)\n",
      ps_threads, ps_make);

  /* create a psolve object */
  a = HaArenaMake();
  st = KheStatsTimerMake(a);
  psolve = PSolveMake(NULL, ins, ps_threads, ps_make, ps_no_diversify, ps_keep,
    ps_time_measure, ps_time_limit, solver, options, a);

  /* do the actual parallel solve */
  PSolveRunAll(psolve);

  /* extract the sole surviving solution */
  HnAssert(HaArrayCount(psolve->instances) == 1,
    "KheInstanceParallelSolve internal error 1");
  pinst = HaArrayFirst(psolve->instances);
  HnAssert(HaArrayCount(pinst->best_solns) == 1,
    "KheInstanceParallelSolve internal error 2");
  soln = HaArrayFirst(pinst->best_solns);
  if( ps_time_measure == KHE_PS_TIME_SHARED )
    KheSolnSetRunningTime(soln, KheStatsTimerNow(st));
  /* PSolveFree(psolve); */
  /* we can't delete this arena, because it contains the arena sets for */
  /* the solutions; if any of those get deleted, the arena sets will be */
  /* called on to receive the recycled arenas */
  /* HaArenaDelete(a); */
  if( DEBUG2 )
  {
    fprintf(stderr, "  final solution:\n");
    KheSolnDebug(soln, 2, 2, stderr);
    fprintf(stderr, "] KheInstanceParallelSolve returning\n");
  }
  return soln;
}
